# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD IFC importer - Enhanced IfcOpenShell-only version"
__author__ = ("Yorik van Havre", "Jonathan Wiedemann", "Bernd Hahnebach")
__url__ = "https://www.freecad.org"

## @package importIFC
#  \ingroup ARCH
#  \brief IFC file format importer
#
#  This module provides tools to import IFC files.

"""Provide the importer for IFC files used above all in Arch and BIM.

Internally it uses IfcOpenShell, which must be installed before using.
"""

import os
import math
import time

import FreeCAD
import Arch
import ArchIFCSchema
import Draft
import DraftVecUtils
import Part

from draftutils import params
from draftutils.messages import _msg, _err

from importers import importIFCHelper
from importers import importIFCmulticore

if FreeCAD.GuiUp:
    import FreeCADGui as Gui

DEBUG = False  # Set to True to see debug messages. Otherwise, totally silent
ZOOMOUT = True  # Set to False to not zoom extents after import

# Templates and other definitions ****
# which IFC type must create which FreeCAD type

typesmap = {
    "Site": ["IfcSite"],
    "Building": ["IfcBuilding"],
    "Floor": ["IfcBuildingStorey"],
    "Structure": [
        "IfcBeam",
        "IfcBeamStandardCase",
        "IfcColumn",
        "IfcColumnStandardCase",
        "IfcSlab",
        "IfcFooting",
        "IfcPile",
        "IfcTendon",
    ],
    "Wall": ["IfcWall", "IfcWallStandardCase", "IfcCurtainWall"],
    "Window": ["IfcWindow", "IfcWindowStandardCase", "IfcDoor", "IfcDoorStandardCase"],
    "Roof": ["IfcRoof"],
    "Stairs": ["IfcStair", "IfcStairFlight", "IfcRamp", "IfcRampFlight"],
    "Space": ["IfcSpace"],
    "Rebar": ["IfcReinforcingBar"],
    "Panel": ["IfcPlate"],
    "Equipment": [
        "IfcFurnishingElement",
        "IfcSanitaryTerminal",
        "IfcFlowTerminal",
        "IfcElectricAppliance",
    ],
    "Pipe": ["IfcPipeSegment"],
    "PipeConnector": ["IfcPipeFitting"],
    "BuildingPart": ["IfcElementAssembly"],
}

# which IFC entity (product) is a structural object
structuralifcobjects = (
    "IfcStructuralCurveMember",
    "IfcStructuralSurfaceMember",
    "IfcStructuralPointConnection",
    "IfcStructuralCurveConnection",
    "IfcStructuralSurfaceConnection",
    "IfcStructuralAction",
    "IfcStructuralPointAction",
    "IfcStructuralLinearAction",
    "IfcStructuralLinearActionVarying",
    "IfcStructuralPlanarAction",
)

# backwards compatibility
getPreferences = importIFCHelper.getPreferences


def export(exportList, filename, colors=None, preferences=None):
    """Export the selected objects to IFC format.

    The export code is now in a separate module; this call is retained
    in this module for compatibility purposes with older scripts.
    """
    from importers import exportIFC

    exportIFC.export(exportList, filename, colors, preferences)


def open(filename, skip=[], only=[], root=None):
    """Open an IFC file inside a new document.

    TODO: change the default argument to `None`, instead of `[]`.
    This is better because lists are mutable.

    Most of the work is done in the `insert` function.
    """
    docname = os.path.splitext(os.path.basename(filename))[0]
    doc = FreeCAD.newDocument(docname)
    doc.Label = docname
    doc = insert(filename, doc.Name, skip, only, root)
    return doc


def insert(srcfile, docname, skip=[], only=[], root=None, preferences=None):
    """Import the contents of an IFC file in the current active document.

    TODO: change the default argument to `None`, instead of `[]`.
    This is better because lists are mutable.

    Parameters
    ----------
    skip: list
        By default empty list `[]`.
        Can contain a list of ids of objects to be skipped.

    only: list
        By default, empty list `[]`.
        Restrict the import to certain object ids; it will also
        get their children.

    root: object
        It is used to import only the derivates of a certain element type,
        for example, `'ifcProduct'
    """
    try:
        import ifcopenshell
        from ifcopenshell import geom

        # Sometimes there is an error importing `geom` in this way
        # import ifcopenshell.geom
        #
        # therefore we must use the `from x import y` way.
        #
        # For some reason this works; see the bug report
        # https://github.com/IfcOpenShell/IfcOpenShell/issues/689
    except ModuleNotFoundError:
        _err(
            "IfcOpenShell was not found on this system. "
            "IFC support is disabled.\n"
            "Visit https://wiki.freecad.org/IfcOpenShell "
            "to learn about installing it."
        )
        return

    starttime = time.time()  # in seconds

    if preferences is None:
        preferences = importIFCHelper.getPreferences()

    if preferences["MULTICORE"] and not hasattr(srcfile, "by_guid"):
        # override with BIM IFC importer if present
        try:
            import BimIfcImport

            return BimIfcImport.insert(srcfile, docname, preferences)
        except:
            pass
        return importIFCmulticore.insert(srcfile, docname, preferences)

    try:
        doc = FreeCAD.getDocument(docname)
    except NameError:
        doc = FreeCAD.newDocument(docname)

    FreeCAD.ActiveDocument = doc

    global parametrics

    # allow one to override the root element
    if root:
        preferences["ROOT_ELEMENT"] = root

    # keeping global variable for debugging purposes
    # global ifcfile

    # If the `by_guid` attribute exists, this is already a loaded ifcfile,
    # otherwise, it's just a string, and we have to open it with ifcopenshell
    if hasattr(srcfile, "by_guid"):
        ifcfile = srcfile
        filesize = None
        filename = None
    else:
        if preferences["DEBUG"]:
            _msg("Opening '{}'... ".format(srcfile), end="")
        filename = srcfile
        filesize = os.path.getsize(filename) * 1e-6  # in megabytes
        ifcfile = ifcopenshell.open(filename)

    if preferences["DEBUG"]:
        _msg("done.")

    # Get file scale
    ifcscale = importIFCHelper.getScaling(ifcfile)

    # IfcOpenShell multiplies the precision value of the file by 100
    # So we raise the precision by 100 too to compensate.
    # ctxs = ifcfile.by_type("IfcGeometricRepresentationContext")
    # for ctx in ctxs:
    #     if not ctx.is_a("IfcGeometricRepresentationSubContext"):
    #         ctx.Precision = ctx.Precision/100

    # Set default ifcopenshell options to work in brep mode
    settings = geom.settings()
    settings.set(settings.USE_BREP_DATA, True)
    settings.set(settings.SEW_SHELLS, True)
    settings.set(settings.USE_WORLD_COORDS, True)
    if preferences["SEPARATE_OPENINGS"]:
        settings.set(settings.DISABLE_OPENING_SUBTRACTIONS, True)
    if preferences["SPLIT_LAYERS"] and hasattr(settings, "APPLY_LAYERSETS"):
        settings.set(settings.APPLY_LAYERSETS, True)

    # build all needed tables
    if preferences["DEBUG"]:
        _msg("Building types and relationships table...")

    # type tables
    sites = ifcfile.by_type("IfcSite")
    buildings = ifcfile.by_type("IfcBuilding")
    floors = ifcfile.by_type("IfcBuildingStorey")
    openings = ifcfile.by_type("IfcOpeningElement")
    materials = ifcfile.by_type("IfcMaterial")
    (products, annotations) = importIFCHelper.buildRelProductsAnnotations(
        ifcfile, preferences["ROOT_ELEMENT"]
    )

    # empty relation tables
    objects = {}  # { id:object, ... }
    shapes = {}  # { id:shaoe } only used for merge mode
    structshapes = {}  # { id:shaoe } only used for merge mode
    sharedobjects = {}  # { representationmapid:object }

    # a list of imported objects whose parametric relationships
    # need processing after all objects have been created
    parametrics = []
    profiles = {}  # to store reused extrusion profiles {ifcid:fcobj, ...}
    layers = {}  # { layer_name, [ids] }
    # filled relation tables

    # TODO: investigate using inverse attributes.
    # For the following tables it might be better to use inverse attributes
    # to find the properties, otherwise a lot of loops
    # and if testing is needed.
    # See https://forum.freecad.org/viewtopic.php?f=39&t=37892
    prodrepr = importIFCHelper.buildRelProductRepresentation(ifcfile)
    additions = importIFCHelper.buildRelAdditions(ifcfile)
    groups = importIFCHelper.buildRelGroups(ifcfile)
    subtractions = importIFCHelper.buildRelSubtractions(ifcfile)
    mattable = importIFCHelper.buildRelMattable(ifcfile)
    colors = importIFCHelper.buildRelProductColors(ifcfile, prodrepr)
    colordict = {}  # { objname:color tuple } for non-GUI use
    if preferences["DEBUG"]:
        _msg("done.")

    # only import a list of IDs and their children, if defined
    if only:
        ids = []
        while only:
            currentid = only.pop()
            ids.append(currentid)
            if currentid in additions:
                only.extend(additions[currentid])
        products = [ifcfile[currentid] for currentid in ids]

    # start the actual import, set FreeCAD UI
    count = 0
    from FreeCAD import Base

    progressbar = Base.ProgressIndicator()
    progressbar.start("Importing IFC objects...", len(products))
    if preferences["DEBUG"]:
        _msg("Parsing {} BIM objects...".format(len(products)))

    # Prepare the 3D view if applicable
    if preferences["FITVIEW_ONIMPORT"] and FreeCAD.GuiUp:
        overallboundbox = None
        Gui.ActiveDocument.activeView().viewAxonometric()

    # Create the base project object
    if not preferences["REPLACE_PROJECT"]:
        if len(ifcfile.by_type("IfcProject")) > 0:
            projectImporter = importIFCHelper.ProjectImporter(ifcfile, objects)
            projectImporter.execute()
        else:
            # https://forum.freecad.org/viewtopic.php?f=39&t=40624
            print("No IfcProject found in the ifc file. Nothing imported")
            return doc

    # handle IFC products

    for product in products:

        count += 1
        pid = product.id()
        guid = product.GlobalId
        ptype = product.is_a()

        if preferences["DEBUG"]:
            print(count, "/", len(products), "object #" + str(pid), ":", ptype, end="")

        # build list of related property sets
        psets = importIFCHelper.getIfcPropertySets(ifcfile, pid)

        # add layer names to layers
        if hasattr(product, "Representation") and hasattr(
            product.Representation, "Representations"
        ):
            if len(product.Representation.Representations) > 0:
                lays = product.Representation.Representations[0].LayerAssignments
                if len(lays) > 0:
                    layer_name = lays[0].Name
                    if layer_name not in layers:
                        layers[layer_name] = [pid]
                    else:
                        layers[layer_name].append(pid)
                    if preferences["DEBUG"]:
                        print(" layer ", layer_name, " found", ptype, end="")
                else:
                    if preferences["DEBUG"]:
                        print(" no layer found", ptype, end="")

        # checking for full FreeCAD parametric definition, overriding everything else
        if psets and params.get_param_arch("IfcImportFreeCADProperties"):
            if "FreeCADPropertySet" in [ifcfile[pset].Name for pset in psets.keys()]:
                if preferences["DEBUG"]:
                    print(" restoring from parametric definition...", end="")
                obj, parametrics = importIFCHelper.createFromProperties(psets, ifcfile, parametrics)
                if obj:
                    objects[pid] = obj
                    if preferences["DEBUG"]:
                        print("done")
                    continue
                else:
                    if preferences["DEBUG"]:
                        print("failed")

        # no parametric data, we go the good old way
        name = str(ptype[3:])
        if product.Name:
            name = product.Name
        if preferences["PREFIX_NUMBERS"]:
            name = "ID" + str(pid) + " " + name
        obj = None
        baseobj = None
        brep = None
        shape = None

        # classify object and verify if we must skip it
        archobj = True  # assume all objects not in structuralifcobjects are architecture
        structobj = False
        if ptype in structuralifcobjects:
            archobj = False
            structobj = True
            if preferences["DEBUG"]:
                print(" (struct)", end="")
        else:
            if preferences["DEBUG"]:
                print(" (arch)", end="")
        if preferences["MERGE_MODE_ARCH"] == 4 and archobj:
            if preferences["DEBUG"]:
                print(" skipped.")
            continue
        if preferences["MERGE_MODE_STRUCT"] == 3 and not archobj:
            if preferences["DEBUG"]:
                print(" skipped.")
            continue
        if pid in skip:  # user given id skip list
            if preferences["DEBUG"]:
                print(" skipped.")
            continue
        if ptype in skip:  # user given type skip list
            if preferences["DEBUG"]:
                print(" skipped.")
            continue
        if ptype in preferences["SKIP"]:  # preferences-set type skip list
            if preferences["DEBUG"]:
                print(" skipped.")
            continue
        if preferences["REPLACE_PROJECT"]:  # options-enabled project/site/building skip
            if ptype in ["IfcProject", "IfcSite"]:
                if preferences["DEBUG"]:
                    print(" skipped.")
                continue
            elif ptype in ["IfcBuilding"]:
                if len(ifcfile.by_type("IfcBuilding")) == 1:
                    # let multiple buildings through...
                    if preferences["DEBUG"]:
                        print(" skipped.")
                    continue
            elif ptype in ["IfcBuildingStorey"]:
                if len(ifcfile.by_type("IfcBuildingStorey")) == 1:
                    # let multiple storeys through...
                    if preferences["DEBUG"]:
                        print(" skipped.")
                    continue

        # check if this object is sharing its shape (mapped representation)
        clone = None
        store = None
        prepr = None
        try:
            prepr = product.Representation
        except Exception:
            if preferences["DEBUG"]:
                print(" ERROR unable to get object representation", end="")
        if (
            prepr
            and (preferences["MERGE_MODE_ARCH"] == 0)
            and archobj
            and preferences["CREATE_CLONES"]
        ):
            for r in prepr.Representations:
                if r.RepresentationIdentifier.upper() == "BODY":
                    if r.Items[0].is_a("IfcMappedItem"):
                        originalid = r.Items[0].MappingSource.id()
                        if originalid in sharedobjects:
                            clone = sharedobjects[originalid]
                        else:
                            sharedobjects[originalid] = None
                            store = originalid  # flag this object to be stored later

        # set additional setting for structural entities
        if hasattr(settings, "INCLUDE_CURVES"):
            if structobj:
                settings.set(settings.INCLUDE_CURVES, True)
            else:
                settings.set(settings.INCLUDE_CURVES, False)
        try:
            cr = geom.create_shape(settings, product)
            brep = cr.geometry.brep_data
        except Exception:
            pass  # IfcOpenShell will yield an error if a given product has no shape, but we don't care, we're brave enough

        # from now on we have a brep string
        if brep:
            if preferences["DEBUG"]:
                print(" " + str(int(len(brep) / 1000)) + "k ", end="")

            # create a Part shape
            shape = Part.Shape()
            shape.importBrepFromString(brep, False)
            shape.scale(
                1000.0
            )  # IfcOpenShell always outputs in meters, we convert to mm, the freecad internal unit

            if shape.isNull() and (not preferences["ALLOW_INVALID"]):
                if preferences["DEBUG"]:
                    print("null shape ", end="")
            elif not shape.isValid() and (not preferences["ALLOW_INVALID"]):
                if preferences["DEBUG"]:
                    print("invalid shape ", end="")
            else:

                # add to the global boundbox if applicable
                if preferences["FITVIEW_ONIMPORT"] and FreeCAD.GuiUp:
                    try:
                        bb = shape.BoundBox
                        # if preferences['DEBUG']: print(' ' + str(bb),end="")
                    except Exception:
                        bb = None
                        if preferences["DEBUG"]:
                            print(" BB could not be computed", end="")
                    if bb and bb.isValid():
                        if not overallboundbox:
                            overallboundbox = bb
                        if not overallboundbox.isInside(bb):
                            Gui.SendMsgToActiveView("ViewFit")
                        overallboundbox.add(bb)

                if (preferences["MERGE_MODE_ARCH"] > 0 and archobj) or structobj:
                    # we are not using Arch objects

                    # additional tweaks to set when not using Arch objects
                    if ptype == "IfcSpace":  # do not add spaces to compounds
                        if preferences["DEBUG"]:
                            print("skipping space ", pid, end="")
                    elif structobj:
                        structshapes[pid] = shape
                        if preferences["DEBUG"]:
                            print(len(shape.Solids), "solids ", end="")
                        baseobj = shape
                    else:
                        shapes[pid] = shape
                        if preferences["DEBUG"]:
                            print(len(shape.Solids), "solids ", end="")
                        baseobj = shape
                else:

                    # create base shape object
                    if clone:
                        if preferences["DEBUG"]:
                            print("clone ", end="")
                    else:
                        if preferences["GET_EXTRUSIONS"] and (preferences["MERGE_MODE_ARCH"] != 1):

                            # get IFC profile
                            profileid = None
                            sortmethod = None
                            if product.Representation:
                                if product.Representation.Representations:
                                    if product.Representation.Representations[0].is_a(
                                        "IfcShapeRepresentation"
                                    ):
                                        if product.Representation.Representations[0].Items:
                                            if (
                                                product.Representation.Representations[0]
                                                .Items[0]
                                                .is_a("IfcExtrudedAreaSolid")
                                            ):
                                                profileid = (
                                                    product.Representation.Representations[0]
                                                    .Items[0]
                                                    .SweptArea.id()
                                                )
                                                sortmethod = importIFCHelper.getProfileCenterPoint(
                                                    product.Representation.Representations[0].Items[
                                                        0
                                                    ]
                                                )

                            # recompose extrusions from a shape
                            if not sortmethod:
                                if ptype in ["IfcWall", "IfcWallStandardCase", "IfcSpace"]:
                                    sortmethod = "z"
                                else:
                                    sortmethod = "area"
                            ex = Arch.getExtrusionData(shape, sortmethod)  # is this an extrusion?
                            if ex:

                                # check for extrusion profile
                                baseface = None
                                addplacement = None
                                if profileid and (profileid in profiles):

                                    # reuse existing profile if existing
                                    print("shared extrusion ", end="")
                                    baseface = profiles[profileid]

                                    # calculate delta placement between stored profile and this one
                                    addplacement = FreeCAD.Placement()
                                    r = FreeCAD.Rotation(
                                        baseface.Shape.Faces[0].normalAt(0, 0),
                                        ex[0].Faces[0].normalAt(0, 0),
                                    )
                                    if r.Angle > 0.000001:
                                        # use shape methods to easily obtain a correct placement
                                        ts = Part.Shape()
                                        ts.rotate(
                                            DraftVecUtils.tup(baseface.Shape.CenterOfMass),
                                            DraftVecUtils.tup(r.Axis),
                                            math.degrees(r.Angle),
                                        )
                                        addplacement = ts.Placement
                                    d = ex[0].CenterOfMass.sub(baseface.Shape.CenterOfMass)
                                    if d.Length > 0.000001:
                                        addplacement.move(d)

                                if not baseface:
                                    # this is an extrusion but we haven't built the profile yet
                                    print("extrusion ", end="")
                                    import DraftGeomUtils

                                    if DraftGeomUtils.hasCurves(ex[0]) or len(ex[0].Wires) != 1:
                                        # is this a circle?
                                        if (len(ex[0].Edges) == 1) and isinstance(
                                            ex[0].Edges[0].Curve, Part.Circle
                                        ):
                                            baseface = Draft.makeCircle(ex[0].Edges[0])
                                        else:
                                            # curves or holes? We just make a Part face
                                            baseface = doc.addObject(
                                                "Part::Feature", name + "_footprint"
                                            )
                                            # bug/feature in ifcopenshell? Some faces of a shell may have non-null placement
                                            # workaround to remove the bad placement: exporting/reimporting as step
                                            if not ex[0].Placement.isNull():
                                                import tempfile

                                                fd, tf = tempfile.mkstemp(suffix=".stp")
                                                ex[0].exportStep(tf)
                                                f = Part.read(tf)
                                                os.close(fd)
                                                os.remove(tf)
                                            else:
                                                f = ex[0]
                                            baseface.Shape = f
                                    else:
                                        # no curve and no hole, we can make a draft object
                                        verts = [v.Point for v in ex[0].Wires[0].OrderedVertexes]
                                        # TODO verts are different if shape is made of RectangleProfileDef or not
                                        # is this a rectangle?
                                        if importIFCHelper.isRectangle(verts):
                                            baseface = Draft.makeRectangle(verts, face=True)
                                        else:
                                            # no hole and no curves, we make a Draft Wire instead
                                            baseface = Draft.makeWire(verts, closed=True)
                                    if profileid:
                                        # store for possible shared use
                                        profiles[profileid] = baseface
                                baseobj = doc.addObject("Part::Extrusion", name + "_body")
                                baseobj.Base = baseface
                                if addplacement:
                                    # apply delta placement (stored profile)
                                    baseobj.Placement = addplacement
                                    baseobj.Dir = addplacement.Rotation.inverted().multVec(ex[1])
                                else:
                                    baseobj.Dir = ex[1]
                                if FreeCAD.GuiUp:
                                    baseface.ViewObject.hide()
                        if not baseobj:
                            baseobj = doc.addObject("Part::Feature", name + "_body")
                            baseobj.Shape = shape
        else:
            # this object has no shape (storeys, etc...)
            if preferences["DEBUG"]:
                print(" no brep ", end="")

        # we now have the shape, we create the final object

        if preferences["MERGE_MODE_ARCH"] == 0 and archobj:

            # full Arch objects

            for freecadtype, ifctypes in typesmap.items():

                if ptype not in ifctypes:
                    continue
                if clone:
                    obj = getattr(Arch, "make" + freecadtype)(name=name)
                    obj.CloneOf = clone

                    # calculate the correct distance from the cloned object
                    if shape:
                        if shape.Solids:
                            s1 = shape.Solids[0]
                        else:
                            s1 = shape
                        if clone.Shape.Solids:
                            s2 = clone.Shape.Solids[0]
                        else:
                            s2 = clone.Shape
                        if hasattr(s1, "CenterOfMass") and hasattr(s2, "CenterOfMass"):
                            v = s1.CenterOfMass.sub(s2.CenterOfMass)
                            if product.Representation:
                                r = importIFCHelper.getRotation(
                                    product.Representation.Representations[0].Items[0].MappingTarget
                                )
                                if not r.isNull():
                                    v = v.add(s2.CenterOfMass)
                                    v = v.add(r.multVec(s2.CenterOfMass.negative()))
                                obj.Placement.Rotation = r
                                obj.Placement.move(v)
                        else:
                            print(
                                "failed to compute placement ",
                            )
                else:
                    # no clone
                    obj = getattr(Arch, "make" + freecadtype)(baseobj=baseobj, name=name)
                    if (
                        freecadtype in ["Wall", "Structure"]
                        and baseobj
                        and baseobj.isDerivedFrom("Part::Extrusion")
                    ):
                        # remove intermediary extrusion for types that can extrude themselves
                        # TODO do this check earlier so we do not calculate it, to save time
                        obj.Base = baseobj.Base
                        obj.Placement = obj.Placement.multiply(baseobj.Placement)
                        obj.Height = baseobj.Dir.Length
                        obj.Normal = FreeCAD.Vector(baseobj.Dir).normalize()
                        bn = baseobj.Name
                        doc.removeObject(bn)
                    if (freecadtype in ["Structure", "Wall"]) and not baseobj:
                        # remove sizes to prevent auto shape creation for types that don't require a base object
                        obj.Height = 0
                        obj.Width = 0
                        obj.Length = 0
                    if (freecadtype in ["Rebar"]) and baseobj:
                        # TODO rebars don't keep link to their base object - we can remove it
                        bn = baseobj.Name
                        doc.removeObject(bn)
                    if store:
                        # Recompute required otherwise obj has a null shape and then
                        # cloning distances cannot be determined in the next loop.
                        obj.recompute()
                        sharedobjects[store] = obj

                # set the placement from the storey's elevation property
                if ptype == "IfcBuildingStorey":
                    if product.Elevation:
                        obj.Placement.Base.z = product.Elevation * ifcscale

                break

            if not obj:
                # we couldn't make an object of a specific type, use default arch component
                obj = Arch.makeComponent(baseobj, name=name)

            # set additional properties
            obj.Label = name
            if preferences["DEBUG"]:
                print(": " + obj.Label + " ", end="")
            if hasattr(obj, "Description") and hasattr(product, "Description"):
                if product.Description:
                    obj.Description = product.Description
            if FreeCAD.GuiUp and baseobj:
                try:
                    if hasattr(baseobj, "ViewObject"):
                        baseobj.ViewObject.hide()
                except ReferenceError:
                    pass

            # setting IFC type

            try:
                if hasattr(obj, "IfcType"):
                    obj.IfcType = "".join(map(lambda x: x if x.islower() else " " + x, ptype[3:]))[
                        1:
                    ]
            except Exception:
                print("Unable to give IFC type ", ptype, " to object ", obj.Label)

            # setting uid

            if hasattr(obj, "IfcData"):
                a = obj.IfcData
                a["IfcUID"] = str(guid)
                obj.IfcData = a

                # setting IFC attributes

                for attribute in ArchIFCSchema.IfcProducts[product.is_a()]["attributes"]:
                    if attribute["name"] == "Name":
                        continue
                    # print("attribute:",attribute["name"])
                    if (
                        hasattr(product, attribute["name"])
                        and getattr(product, attribute["name"])
                        and hasattr(obj, attribute["name"])
                    ):
                        # print("Setting attribute",attribute["name"],"to",getattr(product, attribute["name"]))
                        setattr(obj, attribute["name"], getattr(product, attribute["name"]))
                        # TODO: ArchIFCSchema.IfcProducts uses the IFC version from the FreeCAD prefs.
                        # This might not coincide with the file being opened, hence some attributes are not properly read.

            if obj:
                # print number of solids
                if preferences["DEBUG"]:
                    s = ""
                    if hasattr(obj, "Shape"):
                        if obj.Shape.Solids:
                            s = str(len(obj.Shape.Solids)) + " solids"
                    print(s, end="")
                objects[pid] = obj

        elif (preferences["MERGE_MODE_ARCH"] == 1 and archobj) or (
            preferences["MERGE_MODE_STRUCT"] == 0 and not archobj
        ):

            # non-parametric BIM objects (just Arch components with a shape)

            if ptype in ["IfcSite", "IfcBuilding", "IfcBuildingStorey"]:
                for freecadtype, ifctypes in typesmap.items():
                    if ptype in ifctypes:
                        obj = getattr(Arch, "make" + freecadtype)(baseobj=baseobj, name=name)
                        if preferences["DEBUG"]:
                            print(": " + obj.Label + " ", end="")
                        if ptype == "IfcBuildingStorey":
                            if product.Elevation:
                                obj.Placement.Base.z = product.Elevation * ifcscale
            elif baseobj:
                obj = Arch.makeComponent(baseobj, name=name, delete=True)
                obj.Label = name
                if preferences["DEBUG"]:
                    print(": " + obj.Label + " ", end="")
                if hasattr(obj, "Description") and hasattr(product, "Description"):
                    if product.Description:
                        obj.Description = product.Description
                try:
                    if hasattr(obj, "IfcType"):
                        obj.IfcType = "".join(
                            map(lambda x: x if x.islower() else " " + x, ptype[3:])
                        )[1:]
                except Exception:
                    print("Unable to give IFC type ", ptype, " to object ", obj.Label)
                if hasattr(obj, "IfcData"):
                    a = obj.IfcData
                    a["IfcUID"] = str(guid)
                    obj.IfcData = a
            elif pid in additions:
                # no baseobj but in additions, thus we make a BuildingPart container
                obj = getattr(Arch, "makeBuildingPart")(name=name)
                if preferences["DEBUG"]:
                    print(": " + obj.Label + " ", end="")
            else:
                if preferences["DEBUG"]:
                    print(": skipped.")
                continue
            if obj and hasattr(obj, "GlobalId"):
                obj.GlobalId = guid

        elif (preferences["MERGE_MODE_ARCH"] == 2 and archobj) or (
            preferences["MERGE_MODE_STRUCT"] == 1 and not archobj
        ):

            # Part shapes

            if ptype in ["IfcSite", "IfcBuilding", "IfcBuildingStorey"]:
                for freecadtype, ifctypes in typesmap.items():
                    if ptype in ifctypes:
                        obj = getattr(Arch, "make" + freecadtype)(baseobj=baseobj, name=name)
                        if preferences["DEBUG"]:
                            print(": " + obj.Label + " ", end="")
                        if ptype == "IfcBuildingStorey":
                            if product.Elevation:
                                obj.Placement.Base.z = product.Elevation * ifcscale
            elif baseobj:
                obj = doc.addObject("Part::Feature", name)
                obj.Shape = shape
            elif pid in additions:
                # no baseobj but in additions, thus we make a BuildingPart container
                obj = getattr(Arch, "makeBuildingPart")(name=name)
                if preferences["DEBUG"]:
                    print(": " + obj.Label + " ", end="")
            else:
                if preferences["DEBUG"]:
                    print(": skipped.")
                continue

        if preferences["DEBUG"]:
            print("")  # newline for debug prints, print for a new object should be on a new line

        if obj:

            obj.Label = name
            objects[pid] = obj

            # handle properties

            if psets:

                if preferences["IMPORT_PROPERTIES"] and hasattr(obj, "IfcProperties"):

                    # treat as spreadsheet (pref option)

                    if isinstance(obj.IfcProperties, dict):

                        # fix property type if needed

                        obj.removeProperty("IfcProperties")
                        obj.addProperty(
                            "App::PropertyLink",
                            "IfcProperties",
                            "Component",
                            "Stores IFC properties as a spreadsheet",
                            locked=True,
                        )

                    ifc_spreadsheet = Arch.makeIfcSpreadsheet()
                    n = 2
                    for c in psets.keys():
                        o = ifcfile[c]
                        if preferences["DEBUG"]:
                            print("propertyset Name", o.Name, type(o.Name))
                        catname = o.Name
                        for p in psets[c]:
                            l = ifcfile[p]
                            lname = l.Name
                            if l.is_a("IfcPropertySingleValue"):
                                if preferences["DEBUG"]:
                                    print("property name", l.Name, type(l.Name))
                                ifc_spreadsheet.set(str("A" + str(n)), catname)
                                ifc_spreadsheet.set(str("B" + str(n)), lname)
                                if l.NominalValue:
                                    if preferences["DEBUG"]:
                                        print(
                                            "property NominalValue",
                                            l.NominalValue.is_a(),
                                            type(l.NominalValue.is_a()),
                                        )
                                        print(
                                            "property NominalValue.wrappedValue",
                                            l.NominalValue.wrappedValue,
                                            type(l.NominalValue.wrappedValue),
                                        )
                                        # print("l.NominalValue.Unit",l.NominalValue.Unit,type(l.NominalValue.Unit))
                                    ifc_spreadsheet.set(str("C" + str(n)), l.NominalValue.is_a())
                                    if l.NominalValue.is_a() in [
                                        "IfcLabel",
                                        "IfcText",
                                        "IfcIdentifier",
                                        "IfcDescriptiveMeasure",
                                    ]:
                                        ifc_spreadsheet.set(
                                            str("D" + str(n)),
                                            "'" + str(l.NominalValue.wrappedValue),
                                        )
                                    else:
                                        ifc_spreadsheet.set(
                                            str("D" + str(n)), str(l.NominalValue.wrappedValue)
                                        )
                                    if hasattr(l.NominalValue, "Unit"):
                                        ifc_spreadsheet.set(
                                            str("E" + str(n)), str(l.NominalValue.Unit)
                                        )
                                n += 1
                        obj.IfcProperties = ifc_spreadsheet

                elif hasattr(obj, "IfcProperties") and isinstance(obj.IfcProperties, dict):

                    # 0.18 behaviour: properties are saved as pset;;type;;value in IfcProperties

                    d = obj.IfcProperties
                    obj.IfcProperties = importIFCHelper.getIfcProperties(ifcfile, pid, psets, d)

                elif hasattr(obj, "IfcData"):

                    # 0.17: properties are saved as type(value) in IfcData

                    a = obj.IfcData
                    for c in psets.keys():
                        for p in psets[c]:
                            l = ifcfile[p]
                            if l.is_a("IfcPropertySingleValue"):
                                a[l.Name.encode("utf8")] = str(
                                    l.NominalValue
                                )  # no py3 support here
                    obj.IfcData = a

            # color

            if (pid in colors) and colors[pid]:
                colordict[obj.Name] = colors[pid]
                if FreeCAD.GuiUp:
                    # if preferences['DEBUG']:
                    #    print("    setting color: ",int(colors[pid][0]*255),"/",int(colors[pid][1]*255),"/",int(colors[pid][2]*255))
                    if hasattr(obj.ViewObject, "ShapeColor"):
                        obj.ViewObject.ShapeColor = tuple(colors[pid][0:3])
                    if hasattr(obj.ViewObject, "Transparency"):
                        obj.ViewObject.Transparency = 1.0 - colors[pid][3]

            # if preferences['DEBUG'] is on, recompute after each shape
            if preferences["DEBUG"]:
                doc.recompute()

            # attached 2D elements

            if product.Representation:
                for r in product.Representation.Representations:
                    if r.RepresentationIdentifier == "FootPrint":
                        annotations.append(product)
                        break

            # additional properties for specific types

            if product.is_a("IfcSite"):
                if product.RefElevation:
                    obj.Elevation = product.RefElevation * ifcscale
                if product.RefLatitude:
                    obj.Latitude = importIFCHelper.dms2dd(*product.RefLatitude)
                if product.RefLongitude:
                    obj.Longitude = importIFCHelper.dms2dd(*product.RefLongitude)
                if product.SiteAddress:
                    if product.SiteAddress.AddressLines:
                        obj.Address = product.SiteAddress.AddressLines[0]
                    if product.SiteAddress.Town:
                        obj.City = product.SiteAddress.Town
                    if product.SiteAddress.Region:
                        obj.Region = product.SiteAddress.Region
                    if product.SiteAddress.Country:
                        obj.Country = product.SiteAddress.Country
                    if product.SiteAddress.PostalCode:
                        obj.PostalCode = product.SiteAddress.PostalCode
                project = product.Decomposes[0].RelatingObject
                modelRC = next(
                    (rc for rc in project.RepresentationContexts if rc.ContextType == "Model"), None
                )
                if modelRC and modelRC.TrueNorth:
                    # If the y-part of TrueNorth is 0, then the x-part should be checked.
                    # Declination would be -90° if x  >0 and +90° if x < 0
                    # Only if x==0 then we can not determine TrueNorth.
                    # But that would actually be an invalid IFC file, because the magnitude
                    # of the (twodimensional) direction vector for TrueNorth shall be greater than zero.
                    (x, y) = modelRC.TrueNorth.DirectionRatios[:2]
                    obj.Declination = ((math.degrees(math.atan2(y, x)) - 90 + 180) % 360) - 180
                    if FreeCAD.GuiUp:
                        obj.ViewObject.CompassRotation.Value = obj.Declination

        try:
            progressbar.next(True)
        except RuntimeError:
            print("Aborted.")
            progressbar.stop()
            doc.recompute()
            return

    progressbar.stop()
    doc.recompute()

    if preferences["MERGE_MODE_STRUCT"] == 2:

        if preferences["DEBUG"]:
            print("Joining Structural shapes...", end="")

        for host, children in groups.items():  # Structural
            if ifcfile[host].is_a("IfcStructuralAnalysisModel"):
                compound = []
                for c in children:
                    if c in structshapes:
                        compound.append(structshapes[c])
                        del structshapes[c]
                if compound:
                    name = ifcfile[host].Name or "AnalysisModel"
                    if preferences["PREFIX_NUMBERS"]:
                        name = "ID" + str(host) + " " + name
                    obj = doc.addObject("Part::Feature", name)
                    obj.Label = name
                    obj.Shape = Part.makeCompound(compound)
        if structshapes:  # remaining Structural shapes
            obj = doc.addObject("Part::Feature", "UnclaimedStruct")
            obj.Shape = Part.makeCompound(structshapes.values())

        if preferences["DEBUG"]:
            print("done")

    else:

        if preferences["DEBUG"]:
            print("Processing Struct relationships...", end="")

        # groups

        for host, children in groups.items():
            if ifcfile[host].is_a("IfcStructuralAnalysisModel"):
                # print(host, ' --> ', children)
                obj = doc.addObject("App::DocumentObjectGroup", "AnalysisModel")
                objects[host] = obj
                if host in objects:
                    cobs = []
                    childs_to_delete = []
                    for child in children:
                        if child in objects:
                            cobs.append(objects[child])
                            childs_to_delete.append(child)
                    for c in childs_to_delete:
                        children.remove(c)  # to not process the child again in remaining groups
                    if cobs:
                        if preferences["DEBUG"]:
                            print("adding ", len(cobs), " object(s) to ", objects[host].Label)
                        Arch.addComponents(cobs, objects[host])
                        if preferences["DEBUG"]:
                            doc.recompute()

        if preferences["DEBUG"]:
            print("done")

        if preferences["MERGE_MODE_ARCH"] > 2:  # if ArchObj is compound or ArchObj not imported
            doc.recompute()

            # cleaning bad shapes
            for obj in objects.values():
                if obj.isDerivedFrom("Part::Feature"):
                    if obj.Shape.isNull():
                        Arch.rebuildArchShape(obj)

    # processing remaining (normal) groups

    swallowed = []
    remaining = {}
    for host, children in groups.items():
        if ifcfile[host].is_a("IfcGroup"):
            if ifcfile[host].Name:
                grp_name = ifcfile[host].Name
            else:
                if preferences["DEBUG"]:
                    print(
                        "no group name specified for entity: #",
                        ifcfile[host].id(),
                        ", entity type is used!",
                    )
                grp_name = ifcfile[host].is_a() + "_" + str(ifcfile[host].id())
            grp = doc.addObject("App::DocumentObjectGroup", grp_name)
            grp.Label = grp_name
            objects[host] = grp
            for child in children:
                if child in objects:
                    grp.addObject(objects[child])
                    swallowed.append(child)
                else:
                    remaining[child] = grp

    if preferences["MERGE_MODE_ARCH"] == 3:

        # One compound per storey

        if preferences["DEBUG"]:
            print("Joining Arch shapes...", end="")

        for host, children in additions.items():  # Arch
            if ifcfile[host].is_a("IfcBuildingStorey"):
                compound = []
                for c in children:
                    if c in shapes:
                        compound.append(shapes[c])
                        del shapes[c]
                    if c in additions:
                        for c2 in additions[c]:
                            if c2 in shapes:
                                compound.append(shapes[c2])
                                del shapes[c2]
                if compound:
                    name = ifcfile[host].Name or "Floor"
                    if preferences["PREFIX_NUMBERS"]:
                        name = "ID" + str(host) + " " + name
                    obj = doc.addObject("Part::Feature", name)
                    obj.Label = name
                    obj.Shape = Part.makeCompound(compound)
        if shapes:  # remaining Arch shapes
            obj = doc.addObject("Part::Feature", "UnclaimedArch")
            obj.Shape = Part.makeCompound(shapes.values())

        if preferences["DEBUG"]:
            print("done")

    else:

        if preferences["DEBUG"]:
            print("Processing Arch relationships...", end="")
        first = True

        # subtractions

        if preferences["SEPARATE_OPENINGS"]:
            for subtraction in subtractions:
                if (subtraction[0] in objects) and (subtraction[1] in objects):
                    if preferences["DEBUG"] and first:
                        print("")
                        first = False
                    if preferences["DEBUG"]:
                        print(
                            "    subtracting",
                            objects[subtraction[0]].Label,
                            "from",
                            objects[subtraction[1]].Label,
                        )
                    Arch.removeComponents(objects[subtraction[0]], objects[subtraction[1]])
                    if preferences["DEBUG"]:
                        doc.recompute()

        # additions

        for host, children in additions.items():
            if host not in objects:
                # print(host, 'not used')
                # print(ifcfile[host])
                continue
            cobs = []
            for child in children:
                if (
                    child in objects and child not in swallowed
                ):  # don't add objects already in groups
                    cobs.append(objects[child])
            if not cobs:
                continue
            if preferences["DEBUG"] and first:
                print("")
                first = False
            if (
                preferences["DEBUG"]
                and (len(cobs) > 10)
                and (
                    not (
                        Draft.getType(objects[host])
                        in ["Site", "Building", "Floor", "BuildingPart", "Project"]
                    )
                )
            ):
                # avoid huge fusions
                print("more than 10 shapes to add: skipping.")
            else:
                if preferences["DEBUG"]:
                    print("    adding", len(cobs), "object(s) to", objects[host].Label)
                Arch.addComponents(cobs, objects[host])
                if preferences["DEBUG"]:
                    doc.recompute()

        if preferences["DEBUG"] and first:
            print("done.")

        doc.recompute()

        # cleaning bad shapes

        for obj in objects.values():
            if obj.isDerivedFrom("Part::Feature"):
                if obj.Shape.isNull() and not (Draft.getType(obj) in ["Site", "Project"]):
                    Arch.rebuildArchShape(obj)

    doc.recompute()

    # 2D elements

    if preferences["DEBUG"] and annotations:
        print("Processing", len(annotations), "2D objects...")
    prodcount = count
    count = 0

    for annotation in annotations:

        anno = None
        aid = annotation.id()

        count += 1
        if preferences["DEBUG"]:
            print(
                count, "/", len(annotations), "object #" + str(aid), ":", annotation.is_a(), end=""
            )

        if aid in skip:
            if preferences["DEBUG"]:
                print(", skipped.")
            continue  # user given id skip list
        if annotation.is_a() in preferences["SKIP"]:
            if preferences["DEBUG"]:
                print(", skipped.")
            continue  # preferences-set type skip list

        anno = importIFCHelper.createAnnotation(annotation, doc, ifcscale, preferences)

        # placing in container if needed

        if anno:
            if aid in remaining:
                remaining[aid].addObject(anno)
            else:
                for host, children in additions.items():
                    if (aid in children) and (host in objects):
                        Arch.addComponents(anno, objects[host])

        if preferences["DEBUG"]:
            print("")  # add newline for 2D objects debug prints

    doc.recompute()

    # Materials

    if preferences["DEBUG"] and materials:
        print("Creating materials...", end="")
    # print("\n")
    # print("colors:",colors)
    # print("mattable:",mattable)
    # print("materials:",materials)
    added_mats = []
    for material in materials:
        # print(material.id())

        mdict = {}
        # on ifc import only the "Description" and "DiffuseColor" (if it was read) of the material dictionary will be initialized
        # on editing material in Arch Gui a lot more keys of the material dictionary are initialized even with empty values
        # TODO: there should be a generic material obj init method which will be used by Arch Gui, import IFC, FEM, etc

        # get the material name
        name = "Material"
        if material.Name:
            name = material.Name
        # mdict["Name"] = name on duplicate material names in IFC this could result in crash
        # https://forum.freecad.org/viewtopic.php?f=23&t=63260
        # thus use "Description"
        mdict["Description"] = name

        # get material color
        # the "DiffuseColor" of a material should never be "None"
        # values in colors are None if something went wrong
        # thus the "DiffuseColor" will only be set if the color is not None
        mat_color = None
        if material.id() in colors and colors[material.id()] is not None:
            mat_color = str(colors[material.id()])
        else:
            for o, m in mattable.items():
                if m == material.id():
                    if o in colors and colors[o] is not None:
                        mat_color = str(colors[o])
        if mat_color is not None:
            mdict["DiffuseColor"] = mat_color
        else:
            if preferences["DEBUG"]:
                print("/n  no color for material: {}, ".format(str(material.id)), end="")

        # merge materials with same name and color if setting in prefs is True
        add_material = True
        if preferences["MERGE_MATERIALS"]:
            for added_mat in added_mats:
                if (
                    "Description"
                    in added_mat.Material  # Description has been set thus it is in mdict
                    and added_mat.Material["Description"] == mdict["Description"]
                ):
                    if (
                        (
                            "DiffuseColor" in added_mat.Material
                            and "DiffuseColor" in mdict
                            and added_mat.Material["DiffuseColor"] == mdict["DiffuseColor"]
                        )  # color in added mat with the same matname and new mat is the same
                        or (
                            "DiffuseColor" not in added_mat.Material and "DiffuseColor" not in mdict
                        )  # there is no color in added mat with the same matname and new mat
                        # on model imported from ArchiCAD color was not found for all IFC material objects,
                        # thus DiffuseColor was not set for created materials, workaround to merge these too
                    ):
                        matobj = added_mat
                        add_material = False
                        break

        # add a new material object
        if add_material is True:
            matobj = Arch.makeMaterial(name=name)
            matobj.Material = mdict
            added_mats.append(matobj)
        # fill material attribute of the objects
        for o, m in mattable.items():
            if m == material.id():
                if o in objects:
                    if hasattr(objects[o], "Material"):
                        objects[o].Material = matobj
                        if FreeCAD.GuiUp:
                            # the reason behind ...
                            # there are files around in which the material color is different from the shape color
                            # all viewers use the shape color whereas in FreeCAD the shape color will be
                            # overwritten by the material color (if there is a material with a color).
                            # In such a case FreeCAD shows a different color than all common ifc viewers
                            # https://forum.freecad.org/viewtopic.php?f=39&t=38440
                            col = objects[o].ViewObject.ShapeColor[:3]
                            dig = 5
                            ma_color = sh_color = (
                                round(col[0], dig),
                                round(col[1], dig),
                                round(col[2], dig),
                            )
                            if "DiffuseColor" in objects[o].Material.Material:
                                string_color = objects[o].Material.Material["DiffuseColor"]
                                col = tuple([float(f) for f in string_color.strip("()").split(",")])
                                ma_color = (
                                    round(col[0], dig),
                                    round(col[1], dig),
                                    round(col[2], dig),
                                )
                            if ma_color != sh_color:
                                print("\nobject color != material color for object: ", o)
                                print("    material color is used (most software uses shape color)")
                                print(
                                    "    obj: ", o, "label: ", objects[o].Label, " col: ", sh_color
                                )
                                print("    mat: ", m, "label: ", matobj.Label, " col: ", ma_color)
                                # print("    ", ifcfile[o])
                                # print("    ", ifcfile[m])
                                # print("    colors:")
                                # print("    ", o, ": ", colors[o])
                                # print("    ", m, ": ", colors[m])
    if preferences["DEBUG"] and materials:
        print("done")

    # Grouping everything if required

    # has to be before the Layer
    # if REPLACE_PROJECT and only one storey and one building both are omitted
    # the pure objects do not belong to any container, they will be added here
    # if after Layer they are linked by Layer and will not be added here
    if preferences["REPLACE_PROJECT"] and filename:
        rootgroup = doc.addObject("App::DocumentObjectGroup", "Group")
        rootgroup.Label = os.path.basename(filename)
        # print(objects)
        for key, obj in objects.items():
            # only add top-level objects
            if not obj.InList:
                rootgroup.addObject(obj)

    # Layers

    if preferences["DEBUG"] and layers:
        print("Creating layers...", end="")
    # print(layers)
    for layer_name, layer_objects in layers.items():
        if preferences["IMPORT_LAYER"] is False:
            continue
        # the method make_layer does some nasty debug prints
        lay = Draft.make_layer(layer_name)
        # ShapeColor and LineColor are not set, thus some some default values are used
        # do not override the imported ShapeColor and LineColor with default layer values
        if FreeCAD.GuiUp:
            lay.ViewObject.OverrideLineColorChildren = False
            lay.ViewObject.OverrideShapeColorChildren = False
        lay_grp = []
        for lobj_id in layer_objects:
            if lobj_id in objects:
                lay_grp.append(objects[lobj_id])
        lay.Group = lay_grp
    doc.recompute()
    if preferences["DEBUG"] and layers:
        print("done")

    # restore links from full parametric definitions

    for p in parametrics:
        l = doc.getObject(p[2])
        if l:
            setattr(p[0], p[1], l)

    # Save colordict in non-GUI mode
    if colordict and not FreeCAD.GuiUp:
        import json

        d = doc.Meta
        d["colordict"] = json.dumps(colordict)
        doc.Meta = d

    doc.recompute()

    if FreeCAD.GuiUp and ZOOMOUT:
        Gui.SendMsgToActiveView("ViewFit")

    endtime = time.time() - starttime

    if filesize:
        _msg(
            "Finished importing {0} MB "
            "in {1} seconds, or {2} s/MB".format(
                round(filesize, 1), int(endtime), int(endtime / filesize)
            )
        )
    else:
        _msg("Finished importing in {} seconds".format(int(endtime)))

    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint electrostatic"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"

import FreeCAD


def get_analysis_types():
    return ["electromagnetic"]


def get_sets_name():
    return "constraints_electrostaticpotential_node_sets"


def get_constraint_title():
    return "Fixed electrostatic constraint applied"


def write_meshdata_constraint(f, femobj, pot_obj, ccxwriter):

    if ccxwriter.solver_obj.ElectromagneticMode != "electrostatic":
        return

    if femobj["Object"].BoundaryCondition == "Dirichlet":
        f.write(f"*NSET,NSET={pot_obj.Name}\n")
        for n in femobj["Nodes"]:
            f.write(f"{n},\n")


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def get_before_write_constraint():
    return ""


def get_after_write_constraint():
    return ""


def write_constraint(f, femobj, pot_obj, ccxwriter):

    if ccxwriter.solver_obj.ElectromagneticMode != "electrostatic":
        return

    # floats read from ccx should use {:.13G}, see comment in writer module
    if pot_obj.BoundaryCondition == "Dirichlet":
        f.write("*BOUNDARY\n")
        f.write("{},11,11,{:.13G}\n".format(pot_obj.Name, pot_obj.Potential.getValueAs("mV").Value))
    elif pot_obj.BoundaryCondition == "Neumann":
        density = pot_obj.ElectricFluxDensity.getValueAs("C/mm^2").Value
        # check internal interface
        internal = _check_shared_interface(pot_obj)
        for feat, surf, is_sub_el in femobj["ElectricFluxFaces"]:
            f.write("** {0.Name}.{1[0]}\n".format(*feat))
            f.write("*DFLUX\n")
            d = density
            if feat in internal:
                d = density / 2
            for face, fno in surf:
                f.write("{},S{},{:.13G}\n".format(face, fno, d))

    f.write("\n")


def _check_shared_interface(pot_obj):
    """
    Check if reference is internal shared subshape
    For example, shared face in compsolid
    """
    internal = []
    for o, sub in pot_obj.References:
        for elem in sub:
            found = []
            elem_i = o.getSubObject(elem)
            if elem_i.ShapeType == "Face":
                for s in o.Shape.Solids:
                    found.append(any([q.isSame(elem_i) for q in s.Faces]))
                if sum(found) > 1:
                    internal.append((o, (elem,)))

            if elem_i.ShapeType == "Edge":
                for s in o.Shape.Faces:
                    found.append(any([q.isSame(elem_i) for q in s.Edges]))
                if sum(found) > 1:
                    internal.append((o, (elem,)))

    return internal



# ***************************************************************************
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Python Mesh reader and writer"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package importPyMesh
#  \ingroup FEM
#  \brief FreeCAD Python Mesh reader and writer for FEM workbench

import FreeCAD

from femmesh import meshtools
from builtins import open as pyopen

# ************************************************************************************************
# ********* generic FreeCAD import and export methods ********************************************
# names are fix given from FreeCAD, these methods are called from FreeCAD
# they are set in FEM modules Init.py


# export mesh to python
def export(objectslist, filename):
    "called when freecad exports a file"
    if len(objectslist) != 1:
        FreeCAD.Console.PrintError("This exporter can only export one object.\n")
        return
    obj = objectslist[0]
    if not obj.isDerivedFrom("Fem::FemMeshObject"):
        FreeCAD.Console.PrintError("No FEM mesh object selected.\n")
        return
    femnodes_mesh = obj.FemMesh.Nodes
    femelement_table = meshtools.get_femelement_table(obj.FemMesh)
    if meshtools.is_solid_femmesh(obj.FemMesh):
        fem_mesh_type = "Solid"
    elif meshtools.is_face_femmesh(obj.FemMesh):
        fem_mesh_type = "Face"
    elif meshtools.is_edge_femmesh(obj.FemMesh):
        fem_mesh_type = "Edge"
    else:
        FreeCAD.Console.PrintError("Export of this FEM mesh to Python not supported.\n")
        return
    f = pyopen(filename, "w")
    write_python_mesh_to_file(femnodes_mesh, femelement_table, fem_mesh_type, f)
    f.close()


# ************************************************************************************************
# ********* module specific methods **************************************************************
# writer:
# - a method directly writes a FemMesh to the mesh file
# - a method takes a file handle, mesh data and writes to the file handle

# ********* writer *******************************************************************************


def write(fem_mesh, filename):
    """directly write a FemMesh to a Python mesh file
    fem_mesh: a FemMesh"""

    if not fem_mesh.isDerivedFrom("Fem::FemMesh"):
        FreeCAD.Console.PrintError("Not a FemMesh was given as parameter.\n")
        return
    femnodes_mesh = fem_mesh.Nodes
    femelement_table = meshtools.get_femelement_table(fem_mesh)
    if meshtools.is_solid_femmesh(fem_mesh):
        fem_mesh_type = "Solid"
    elif meshtools.is_face_femmesh(fem_mesh):
        fem_mesh_type = "Face"
    elif meshtools.is_edge_femmesh(fem_mesh):
        fem_mesh_type = "Edge"
    else:
        FreeCAD.Console.PrintError("Export of this FEM mesh to Python not supported.\n")
        return
    f = pyopen(filename, "w")
    write_python_mesh_to_file(femnodes_mesh, femelement_table, fem_mesh_type, f)
    f.close()


def write_python_mesh_to_file(femnodes_mesh, femelement_table, fem_mesh_type, f):

    mesh_name = "femmesh"

    # nodes
    f.write("def create_nodes(femmesh):\n")
    f.write("    # nodes\n")
    for node in femnodes_mesh:
        # print(node, ' --> ', femnodes_mesh[node])
        vec = femnodes_mesh[node]
        f.write(f"    {mesh_name}.addNode({vec.x}, {vec.y}, {vec.z}, {node})\n")
    f.write("    return True\n")
    f.write("\n\n")

    # elements
    f.write("def create_elements(femmesh):\n")
    f.write("    # elements\n")
    for element in femelement_table:
        # print(element, ' --> ', femelement_table[element])
        if fem_mesh_type == "Solid":
            f.write(
                "    {}.addVolume({}, {})\n".format(
                    mesh_name, list(femelement_table[element]), element
                )
            )
        elif fem_mesh_type == "Face":
            f.write(
                "    {}.addFace({}, {})\n".format(
                    mesh_name, list(femelement_table[element]), element
                )
            )
        elif fem_mesh_type == "Edge":
            f.write(
                "    {}.addEdge({}, {})\n".format(
                    mesh_name, list(femelement_table[element]), element
                )
            )
    f.write("    return True\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2023 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# **************************************************************************/

import FreeCAD
import unittest


class TestStringHasher(unittest.TestCase):
    def setUp(self):
        self.strHash = FreeCAD.StringHasher()
        self.strID = self.strHash.getID("A")

    def testInit(self):
        with self.assertRaises(TypeError):
            FreeCAD.StringHasher(0)

    def testGetID(self):
        with self.assertRaises(ValueError):
            self.strHash.getID(0)

    def testStringHasherIsSame(self):
        with self.assertRaises(TypeError):
            self.strHash.isSame(0)

    def testStringIDIsSame(self):
        with self.assertRaises(TypeError):
            self.strID.isSame(0)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""This is the main NativeIFC module"""

import os

from PySide import QtCore

import FreeCAD
import Arch
import ArchBuildingPart
import Draft

from draftviewproviders import view_layer

translate = FreeCAD.Qt.translate

# heavyweight libraries - ifc_tools should always be lazy loaded

try:
    import ifcopenshell
    import ifcopenshell.api
    import ifcopenshell.geom
    import ifcopenshell.util.attribute
    import ifcopenshell.util.element
    import ifcopenshell.util.placement
    import ifcopenshell.util.schema
    import ifcopenshell.util.unit
    import ifcopenshell.entity_instance
except ImportError as e:
    import FreeCAD

    FreeCAD.Console.PrintError(
        translate(
            "BIM",
            "IfcOpenShell was not found on this system. IFC support is disabled",
        )
        + "\n"
    )
    raise e

from . import ifc_objects
from . import ifc_viewproviders
from . import ifc_import
from . import ifc_layers
from . import ifc_status
from . import ifc_export
from . import ifc_psets

SCALE = 1000.0  # IfcOpenShell works in meters, FreeCAD works in mm
SHORT = False  # If True, only Step ID attribute is created
ROUND = 8  # rounding value for placements
DEFAULT_SHAPEMODE = "Coin"  # Can be Shape, Coin or None
PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/NativeIFC")


def create_document(document, filename=None, shapemode=0, strategy=0, silent=False):
    """Creates a IFC document object in the given FreeCAD document or converts that
    document into an IFC document, depending on the state of the statusbar lock button.

    filename:  If not given, a blank IFC document is created
    shapemode: 0 = full shape
               1 = coin only
               2 = no representation
    strategy:  0 = only root object
               1 = only building structure
               2 = all children
    """

    if ifc_status.get_lock_status():
        return convert_document(document, filename, shapemode, strategy, silent)
    else:
        return create_document_object(document, filename, shapemode, strategy, silent)


def create_document_object(document, filename=None, shapemode=0, strategy=0, silent=False):
    """Creates a IFC document object in the given FreeCAD document.

    filename:  If not given, a blank IFC document is created
    shapemode: 0 = full shape
               1 = coin only
               2 = no representation
    strategy:  0 = only root object
               1 = only building structure
               2 = all children
    """

    obj = add_object(document, otype="project")
    ifcfile, project, full = setup_project(obj, filename, shapemode, silent)
    # populate according to strategy
    if strategy == 0:
        pass
    elif strategy == 1:
        create_children(obj, ifcfile, recursive=True, only_structure=True)
    elif strategy == 2:
        create_children(obj, ifcfile, recursive=True, assemblies=False)
    # create default structure
    if full:
        site = aggregate(Arch.makeSite(), obj)
        building = aggregate(Arch.makeBuilding(), site)
        storey = aggregate(Arch.makeFloor(), building)
    return obj


def convert_document(document, filename=None, shapemode=0, strategy=0, silent=False):
    """Converts the given FreeCAD document to an IFC document.

    filename:  If not given, a blank IFC document is created
    shapemode: 0 = full shape
               1 = coin only
               2 = no representation
    strategy:  0 = only root object
               1 = only bbuilding structure
               2 = all children
               3 = no children
    """

    if "Proxy" not in document.PropertiesList:
        document.addProperty("App::PropertyPythonObject", "Proxy", locked=True)
    document.setPropertyStatus("Proxy", "Transient")
    document.Proxy = ifc_objects.document_object()
    ifcfile, project, full = setup_project(document, filename, shapemode, silent)
    if strategy == 0:
        create_children(document, ifcfile, recursive=False)
    elif strategy == 1:
        create_children(document, ifcfile, recursive=True, only_structure=True)
    elif strategy == 2:
        create_children(document, ifcfile, recursive=True, assemblies=False)
    elif strategy == 3:
        pass
    # create default structure
    if full:
        site = aggregate(Arch.makeSite(), document)
        building = aggregate(Arch.makeBuilding(), site)
        storey = aggregate(Arch.makeFloor(), building)
    return document


def setup_project(proj, filename, shapemode, silent):
    """Sets up a project (common operations between single doc/not single doc modes)
    Returns the ifcfile object, the project ifc entity, and full (True/False)"""

    full = False
    d = "The path to the linked IFC file"
    if "IfcFilePath" not in proj.PropertiesList:
        proj.addProperty("App::PropertyFile", "IfcFilePath", "Base", d, locked=True)
    if "Modified" not in proj.PropertiesList:
        proj.addProperty("App::PropertyBool", "Modified", "Base", locked=True)
    proj.setPropertyStatus("Modified", "Hidden")
    if filename:
        # opening existing file
        proj.IfcFilePath = filename
        ifcfile = ifcopenshell.open(filename)
    else:
        # creating a new file
        if not silent:
            full = ifc_import.get_project_type()
        ifcfile = create_ifcfile()
    project = ifcfile.by_type("IfcProject")[0]
    # TODO configure version history
    # https://blenderbim.org/docs-python/autoapi/ifcopenshell/api/owner/create_owner_history/index.html
    # In IFC4, history is optional. What should we do here?
    proj.Proxy.ifcfile = ifcfile
    add_properties(proj, ifcfile, project, shapemode=shapemode)
    if "Schema" not in proj.PropertiesList:
        proj.addProperty("App::PropertyEnumeration", "Schema", "Base", locked=True)
    # bug in FreeCAD - to avoid a crash, pre-populate the enum with one value
    proj.Schema = [ifcfile.wrapped_data.schema_name()]
    proj.Schema = ifcfile.wrapped_data.schema_name()
    proj.Schema = ifcopenshell.ifcopenshell_wrapper.schema_names()
    return ifcfile, project, full


def create_ifcfile():
    """Creates a new, empty IFC document"""

    ifcfile = api_run("project.create_file")
    project = api_run("root.create_entity", ifcfile, ifc_class="IfcProject")
    param = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Document")
    user = param.GetString("prefAuthor", "")
    user = user.split("<")[0].strip()
    org = param.GetString("prefCompany", "")
    person = None
    organisation = None
    if user:
        person = api_run("owner.add_person", ifcfile, family_name=user)
    if org:
        organisation = api_run("owner.add_organisation", ifcfile, name=org)
    if user and org:
        api_run(
            "owner.add_person_and_organisation",
            ifcfile,
            person=person,
            organisation=organisation,
        )
    application = "FreeCAD"
    version = FreeCAD.Version()
    version = ".".join([str(v) for v in version[0:3]])
    freecadorg = api_run(
        "owner.add_organisation", ifcfile, identification="FreeCAD.org", name="The FreeCAD project"
    )
    application = api_run(
        "owner.add_application",
        ifcfile,
        application_developer=freecadorg,
        application_full_name=application,
        application_identifier=application,
        version=version,
    )
    # context
    model3d = api_run("context.add_context", ifcfile, context_type="Model")
    plan = api_run("context.add_context", ifcfile, context_type="Plan")
    body = api_run(
        "context.add_context",
        ifcfile,
        context_type="Model",
        context_identifier="Body",
        target_view="MODEL_VIEW",
        parent=model3d,
    )
    api_run(
        "context.add_context",
        ifcfile,
        context_type="Model",
        context_identifier="Axis",
        target_view="GRAPH_VIEW",
        parent=model3d,
    )
    # unit
    # for now, assign a default metre + sqm +degrees unit, as per
    # https://docs.ifcopenshell.org/autoapi/ifcopenshell/api/unit/index.html
    # TODO allow to set this at creation, from the current FreeCAD units schema
    length = api_run("unit.add_si_unit", ifcfile, unit_type="LENGTHUNIT")
    area = api_run("unit.add_si_unit", ifcfile, unit_type="AREAUNIT")
    angle = api_run("unit.add_conversion_based_unit", ifcfile, name="degree")
    api_run("unit.assign_unit", ifcfile, units=[length, area, angle])
    # TODO add user history
    return ifcfile


def api_run(*args, **kwargs):
    """Runs an IfcOpenShell API call and flags the ifcfile as modified"""

    result = ifcopenshell.api.run(*args, **kwargs)
    # *args are typically command, ifcfile
    if len(args) > 1:
        ifcfile = args[1]
        for d in FreeCAD.listDocuments().values():
            for o in d.Objects:
                if hasattr(o, "Proxy") and hasattr(o.Proxy, "ifcfile"):
                    if o.Proxy.ifcfile == ifcfile:
                        o.Modified = True
    return result


def create_object(ifcentity, document, ifcfile, shapemode=0, objecttype=None):
    """Creates a FreeCAD object from an IFC entity"""

    exobj = get_object(ifcentity, document)
    if exobj:
        return exobj
    s = "IFC: Created #{}: {}, '{}'\n".format(
        ifcentity.id(), ifcentity.is_a(), getattr(ifcentity, "Name", "")
    )
    objecttype = ifc_export.get_object_type(ifcentity, objecttype)
    FreeCAD.Console.PrintLog(s)
    obj = add_object(document, otype=objecttype)
    add_properties(obj, ifcfile, ifcentity, shapemode=shapemode)
    ifc_layers.add_layers(obj, ifcentity, ifcfile)
    if FreeCAD.GuiUp:
        if (
            ifcentity.is_a("IfcSpace")
            or ifcentity.is_a("IfcOpeningElement")
            or ifcentity.is_a("IfcAnnotation")
        ):
            try:
                obj.ViewObject.DisplayMode = "Wireframe"
            except:
                pass
    elements = [ifcentity]
    return obj


def create_children(
    obj,
    ifcfile=None,
    recursive=False,
    only_structure=False,
    assemblies=True,
    expand=False,
):
    """Creates a hierarchy of objects under an object"""

    def get_parent_objects(parent):
        proj = get_project(parent)
        if hasattr(proj, "OutListRecursive"):
            return proj.OutListRecursive
        elif hasattr(proj, "Objects"):
            return proj.Objects

    def create_child(parent, element):
        subresult = []
        # do not create if a child with same stepid already exists
        if element.id() not in [getattr(c, "StepId", 0) for c in get_parent_objects(parent)]:
            doc = getattr(parent, "Document", parent)
            mode = getattr(parent, "ShapeMode", "Coin")
            child = create_object(element, doc, ifcfile, mode)
            subresult.append(child)
            if isinstance(parent, FreeCAD.DocumentObject):
                parent.Proxy.addObject(parent, child)
            if element.is_a("IfcSite"):
                # force-create contained buildings too if we just created a site
                buildings = [o for o in get_children(child, ifcfile) if o.is_a("IfcBuilding")]
                for building in buildings:
                    subresult.extend(create_child(child, building))
            elif element.is_a("IfcOpeningElement"):
                # force-create contained windows too if we just created an opening
                windows = [
                    o for o in get_children(child, ifcfile) if o.is_a() in ("IfcWindow", "IfcDoor")
                ]
                for window in windows:
                    subresult.extend(create_child(child, window))

            if recursive:
                subresult.extend(
                    create_children(child, ifcfile, recursive, only_structure, assemblies)
                )
        return subresult

    if not ifcfile:
        ifcfile = get_ifcfile(obj)
    result = []
    children = get_children(obj, ifcfile, only_structure, assemblies, expand)
    for child in children:
        result.extend(create_child(obj, child))
    assign_groups(children)
    # TEST: mark new objects to recompute
    QtCore.QTimer.singleShot(0, lambda: recompute([get_object(c) for c in children]))
    return result


def assign_groups(children, ifcfile=None):
    """Fill the groups in this list. Returns a list of processed FreeCAD objects"""

    result = []
    for child in children:
        if child.is_a("IfcGroup"):
            mode = "IsGroupedBy"
        elif child.is_a("IfcElementAssembly"):
            mode = "IsDecomposedBy"
        else:
            mode = None
        if mode:
            grobj = get_object(child, None, ifcfile)
            for rel in getattr(child, mode):
                for elem in rel.RelatedObjects:
                    elobj = get_object(elem, None, ifcfile)
                    if elobj:
                        if len(elobj.InList) == 1:
                            p = elobj.InList[0]
                            if elobj in p.Group:
                                g = p.Group
                                g.remove(elobj)
                                p.Group = g
                        g = grobj.Group
                        g.append(elobj)
                        grobj.Group = g
                        result.append(elobj)
    return result


def get_children(
    obj, ifcfile=None, only_structure=False, assemblies=True, expand=False, ifctype=None
):
    """Returns the direct descendants of an object"""

    if not ifcfile:
        ifcfile = get_ifcfile(obj)
    ifcentity = ifcfile[obj.StepId]
    children = []
    if assemblies or not ifcentity.is_a("IfcElement"):
        for rel in getattr(ifcentity, "IsDecomposedBy", []):
            children.extend(rel.RelatedObjects)
    if not only_structure:
        for rel in getattr(ifcentity, "ContainsElements", []):
            children.extend(rel.RelatedElements)
        for rel in getattr(ifcentity, "HasOpenings", []):
            children.extend([rel.RelatedOpeningElement])
        for rel in getattr(ifcentity, "HasFillings", []):
            children.extend([rel.RelatedBuildingElement])
    result = filter_elements(children, ifcfile, expand=expand, spaces=True, assemblies=assemblies)
    if ifctype:
        result = [r for r in result if r.is_a(ifctype)]
    return result


def get_freecad_children(obj):
    """Returns the children of this object that exist in the document"""

    objs = []
    children = get_children(obj)
    for child in children:
        childobj = get_object(child)
        if childobj:
            objs.extend(get_freecad_children(childobj))
    return objs


def get_object(element, document=None, ifcfile=None):
    """Returns the object that references this element, if any"""

    if document:
        ldocs = {"document": document}
    else:
        ldocs = FreeCAD.listDocuments()
    for n, d in ldocs.items():
        for obj in d.Objects:
            if hasattr(obj, "StepId"):
                if obj.StepId == element.id():
                    if get_ifc_element(obj, ifcfile) == element:
                        return obj
    return None


def get_ifcfile(obj):
    """Returns the ifcfile that handles this object"""

    project = get_project(obj)
    if project:
        if getattr(project, "Proxy", None):
            if hasattr(project.Proxy, "ifcfile"):
                return project.Proxy.ifcfile
        if getattr(project, "IfcFilePath", None):
            ifcfile = ifcopenshell.open(project.IfcFilePath)
            if hasattr(project, "Proxy"):
                if project.Proxy is None:
                    if not isinstance(project, FreeCAD.DocumentObject):
                        project.Proxy = ifc_objects.document_object()
            if getattr(project, "Proxy", None):
                project.Proxy.ifcfile = ifcfile
            return ifcfile
        else:
            FreeCAD.Console.PrintError(
                "Error: No IFC file attached to this project: " + project.Label
            )
    return None


def get_project(obj):
    """Returns the ifc document this object belongs to.
    obj can be either a document object, an ifcfile or ifc element instance"""

    proj_types = ("IfcProject", "IfcProjectLibrary")
    if isinstance(obj, ifcopenshell.file):
        for d in FreeCAD.listDocuments().values():
            for o in d.Objects:
                if hasattr(o, "Proxy") and hasattr(o.Proxy, "ifcfile"):
                    if o.Proxy.ifcfile == obj:
                        return o
        return None
    if isinstance(obj, ifcopenshell.entity_instance):
        obj = get_object(obj)
    if hasattr(obj, "IfcFilePath"):
        return obj
    if hasattr(getattr(obj, "Document", None), "IfcFilePath"):
        return obj.Document
    if getattr(obj, "Class", None) in proj_types:
        return obj
    if hasattr(obj, "InListRecursive"):
        for parent in obj.InListRecursive:
            if getattr(parent, "Class", None) in proj_types:
                return parent
    return None


def can_expand(obj, ifcfile=None):
    """Returns True if this object can have any more child extracted"""

    if not ifcfile:
        ifcfile = get_ifcfile(obj)
    children = get_children(obj, ifcfile, expand=True)
    group = [o.StepId for o in obj.Group if hasattr(o, "StepId")]
    for child in children:
        if child.id() not in group:
            return True
    return False


def add_object(document, otype=None, oname="IfcObject"):
    """adds a new object to a FreeCAD document.
    otype can be:
    'project',
    'group',
    'material',
    'layer',
    'text',
    'dimension',
    'sectionplane',
    'axis',
    'schedule'
    'buildingpart'
    or anything else for a standard IFC object"""

    if not document:
        return None
    if otype == "schedule":
        obj = Arch.makeSchedule()
    elif otype == "sectionplane":
        obj = Arch.makeSectionPlane()
        obj.Proxy = ifc_objects.ifc_object(otype)
    elif otype == "axis":
        obj = Arch.makeAxis()
        obj.Proxy = ifc_objects.ifc_object(otype)
        obj.removeProperty("Angles")
        obj.removeProperty("Distances")
        obj.removeProperty("Labels")
        obj.removeProperty("Limit")
        if obj.ViewObject:
            obj.ViewObject.DisplayMode = "Flat Lines"
    elif otype == "dimension":
        obj = Draft.make_dimension(FreeCAD.Vector(), FreeCAD.Vector(1, 0, 0))
        obj.Proxy = ifc_objects.ifc_object(otype)
        obj.removeProperty("Diameter")
        obj.removeProperty("Distance")
        obj.setPropertyStatus("LinkedGeometry", "Hidden")
        obj.setGroupOfProperty("Start", "Dimension")
        obj.setGroupOfProperty("End", "Dimension")
        obj.setGroupOfProperty("Direction", "Dimension")
    elif otype == "text":
        obj = Draft.make_text("")
        obj.Proxy = ifc_objects.ifc_object(otype)
    elif otype == "layer":
        proxy = ifc_objects.ifc_object(otype)
        obj = document.addObject("App::FeaturePython", oname, proxy, None, False)
        if obj.ViewObject:
            view_layer.ViewProviderLayer(obj.ViewObject)
            obj.ViewObject.addProperty("App::PropertyBool", "HideChildren", "Layer", locked=True)
            obj.ViewObject.HideChildren = True
    elif otype == "group":
        vproxy = ifc_viewproviders.ifc_vp_group()
        obj = document.addObject("App::DocumentObjectGroupPython", oname, None, vproxy, False)
    elif otype == "material":
        proxy = ifc_objects.ifc_object(otype)
        vproxy = ifc_viewproviders.ifc_vp_material()
        obj = document.addObject("App::MaterialObjectPython", oname, proxy, vproxy, False)
    elif otype == "project":
        proxy = ifc_objects.ifc_object(otype)
        vproxy = ifc_viewproviders.ifc_vp_document()
        obj = document.addObject("Part::FeaturePython", oname, proxy, vproxy, False)
    elif otype == "buildingpart":
        obj = Arch.makeBuildingPart()
        if obj.ViewObject:
            obj.ViewObject.ShowLevel = False
            obj.ViewObject.ShowLabel = False
            obj.ViewObject.Proxy = ifc_viewproviders.ifc_vp_buildingpart(obj.ViewObject)
            obj.ViewObject.Proxy.attach(obj.ViewObject)
        for p in obj.PropertiesList:
            if obj.getGroupOfProperty(p) in ["BuildingPart", "IFC Attributes", "Children"]:
                obj.removeProperty(p)
        obj.Proxy = ifc_objects.ifc_object(otype)
    else:  # default case, standard IFC object
        proxy = ifc_objects.ifc_object(otype)
        vproxy = ifc_viewproviders.ifc_vp_object()
        obj = document.addObject("Part::FeaturePython", oname, proxy, vproxy, False)
    return obj


def add_properties(obj, ifcfile=None, ifcentity=None, links=False, shapemode=0, short=SHORT):
    """Adds the properties of the given IFC object to a FreeCAD object"""

    if not ifcfile:
        ifcfile = get_ifcfile(obj)
    if not ifcentity:
        ifcentity = get_ifc_element(obj)
    if getattr(ifcentity, "Name", None):
        obj.Label = ifcentity.Name
    elif getattr(obj, "IfcFilePath", ""):
        obj.Label = os.path.splitext(os.path.basename(obj.IfcFilePath))[0]
    else:
        obj.Label = "_" + ifcentity.is_a()
    if isinstance(obj, FreeCAD.DocumentObject) and "Group" not in obj.PropertiesList:
        obj.addProperty("App::PropertyLinkList", "Group", "Base", locked=True)
    if "ShapeMode" not in obj.PropertiesList:
        obj.addProperty("App::PropertyEnumeration", "ShapeMode", "Base", locked=True)
        shapemodes = [
            "Shape",
            "Coin",
            "None",
        ]  # possible shape modes for all IFC objects
        if isinstance(shapemode, int):
            shapemode = shapemodes[shapemode]
        obj.ShapeMode = shapemodes
        obj.ShapeMode = shapemode
        if not obj.isDerivedFrom("Part::Feature"):
            obj.setPropertyStatus("ShapeMode", "Hidden")
    if ifcentity.is_a("IfcProduct"):
        obj.addProperty("App::PropertyLink", "Type", "IFC", locked=True)
    attr_defs = ifcentity.wrapped_data.declaration().as_entity().all_attributes()
    try:
        info_ifcentity = ifcentity.get_info()
    except:
        # slower but no errors
        info_ifcentity = get_elem_attribs(ifcentity)
    for attr, value in info_ifcentity.items():
        if attr == "type":
            attr = "Class"
        elif attr == "id":
            attr = "StepId"
        elif attr == "Name":
            continue
        if short and attr not in ("Class", "StepId"):
            continue
        attr_def = next((a for a in attr_defs if a.name() == attr), None)
        data_type = ifcopenshell.util.attribute.get_primitive_type(attr_def) if attr_def else None
        if attr == "Class":
            # main enum property, not saved to file
            if attr not in obj.PropertiesList:
                obj.addProperty("App::PropertyEnumeration", attr, "IFC", locked=True)
                obj.setPropertyStatus(attr, "Transient")
            # to avoid bug/crash: we populate first the property with only the
            # class, then we add the sibling classes
            setattr(obj, attr, [value])
            setattr(obj, attr, value)
            setattr(obj, attr, get_ifc_classes(obj, value))
            # companion hidden propertym that gets saved to file
            if "IfcClass" not in obj.PropertiesList:
                obj.addProperty("App::PropertyString", "IfcClass", "IFC", locked=True)
                obj.setPropertyStatus("IfcClass", "Hidden")
            setattr(obj, "IfcClass", value)
        elif attr_def and "IfcLengthMeasure" in str(attr_def.type_of_attribute()):
            obj.addProperty("App::PropertyDistance", attr, "IFC")
            if value:
                setattr(obj, attr, value * (1 / get_scale(ifcfile)))
        elif isinstance(value, int):
            if attr not in obj.PropertiesList:
                obj.addProperty("App::PropertyInteger", attr, "IFC", locked=True)
                if attr == "StepId":
                    obj.setPropertyStatus(attr, "ReadOnly")
            setattr(obj, attr, value)
        elif isinstance(value, float):
            if attr not in obj.PropertiesList:
                obj.addProperty("App::PropertyFloat", attr, "IFC", locked=True)
            setattr(obj, attr, value)
        elif data_type == "boolean":
            if attr not in obj.PropertiesList:
                obj.addProperty("App::PropertyBool", attr, "IFC", locked=True)
            if not value or value in ["UNKNOWN", "FALSE"]:
                value = False
            elif not isinstance(value, bool):
                print("DEBUG: attempting to set boolean value:", attr, value)
                value = bool(value)
            setattr(obj, attr, value)  # will trigger error. TODO: Fix this
        elif isinstance(value, ifcopenshell.entity_instance):
            if links:
                if attr not in obj.PropertiesList:
                    obj.addProperty("App::PropertyLink", attr, "IFC", locked=True)
        elif isinstance(value, (list, tuple)) and value:
            if isinstance(value[0], ifcopenshell.entity_instance):
                if links:
                    if attr not in obj.PropertiesList:
                        obj.addProperty("App::PropertyLinkList", attr, "IFC", locked=True)
        elif data_type == "enum":
            if attr not in obj.PropertiesList:
                obj.addProperty("App::PropertyEnumeration", attr, "IFC", locked=True)
            items = ifcopenshell.util.attribute.get_enum_items(attr_def)
            if value not in items:
                for v in ("UNDEFINED", "NOTDEFINED", "USERDEFINED"):
                    if v in items:
                        value = v
                        break
            if value in items:
                # to prevent bug/crash, we first need to populate the
                # enum with the value about to be used, then
                # add the alternatives
                setattr(obj, attr, [value])
                setattr(obj, attr, value)
                setattr(obj, attr, items)
        elif attr in ["RefLongitude", "RefLatitude"]:
            obj.addProperty("App::PropertyFloat", attr, "IFC", locked=True)
            if value is not None:
                # convert from list of 4 ints
                value = value[0] + value[1] / 60.0 + value[2] / 3600.0 + value[3] / 3600.0e6
                setattr(obj, attr, value)
        else:
            if attr not in obj.PropertiesList:
                obj.addProperty("App::PropertyString", attr, "IFC", locked=True)
            if value is not None:
                setattr(obj, attr, str(value))

            # We shortly go through the list of IFCRELASSOCIATESCLASSIFICATION members
            # in the file to see if the newly added object should have a Classification added
            # since we can run `add_properties`, when changing from IFC Object to IFC Type, or BIM Object (Standard Code)
            # to BIM Type, and during the process of creation the only place where we save Classification is
            # the file itself, so below code retrieves it and assigns it back to the newly created obj.
            if not hasattr(obj, "Classification"):
                assoc_classifications = ifcfile.by_type("IfcRelAssociatesClassification")
                for assoc in assoc_classifications:
                    related_objects = assoc.RelatedObjects
                    if isinstance(related_objects, ifcopenshell.entity_instance):
                        related_objects = [related_objects]
                    if ifcentity in related_objects:
                        cref = assoc.RelatingClassification
                        if cref and cref.is_a("IfcClassificationReference"):
                            classification_name = ""

                            # Try to get the source classification name
                            if hasattr(cref, "ReferencedSource") and cref.ReferencedSource:
                                if (
                                    hasattr(cref.ReferencedSource, "Name")
                                    and cref.ReferencedSource.Name
                                ):
                                    classification_name += cref.ReferencedSource.Name + " "

                            # Add the Identification if present
                            if cref.Identification:
                                classification_name += cref.Identification

                            classification_name = classification_name.strip()
                            if classification_name:
                                obj.addProperty(
                                    "App::PropertyString", "Classification", "IFC", locked=True
                                )
                                setattr(obj, "Classification", classification_name)
                                break  # Found the relevant one, stop
    # annotation properties
    if ifcentity.is_a("IfcGridAxis"):
        axisdata = ifc_export.get_axis(ifcentity)
        if axisdata:
            if "Placement" not in obj.PropertiesList:
                obj.addProperty("App::PropertyPlacement", "Placement", "Base", locked=True)
            if "CustomText" in obj.PropertiesList:
                obj.setPropertyStatus("CustomText", "Hidden")
                obj.setExpression("CustomText", "AxisTag")
            if "Length" not in obj.PropertiesList:
                obj.addProperty("App::PropertyLength", "Length", "Axis", locked=True)
            if "Text" not in obj.PropertiesList:
                obj.addProperty("App::PropertyStringList", "Text", "Base", locked=True)
            obj.Placement = axisdata[0]
            obj.Length = axisdata[1]
            # axisdata[2] is the axis tag, it is already applied by other code
    elif ifcentity.is_a("IfcAnnotation"):
        sectionplane = ifc_export.get_sectionplane(ifcentity)
        if sectionplane:
            if "Placement" not in obj.PropertiesList:
                obj.addProperty("App::PropertyPlacement", "Placement", "Base", locked=True)
            if "Depth" not in obj.PropertiesList:
                obj.addProperty("App::PropertyLength", "Depth", "SectionPlane", locked=True)
            obj.Placement = sectionplane[0]
            if len(sectionplane) > 3:
                obj.Depth = sectionplane[3]
            vobj = obj.ViewObject
            if vobj:
                if "DisplayLength" not in vobj.PropertiesList:
                    vobj.addProperty(
                        "App::PropertyLength", "DisplayLength", "SectionPlane", locked=True
                    )
                if "DisplayHeight" not in vobj.PropertiesList:
                    vobj.addProperty(
                        "App::PropertyLength", "DisplayHeight", "SectionPlane", locked=True
                    )
                if len(sectionplane) > 1:
                    vobj.DisplayLength = sectionplane[1]
                if len(sectionplane) > 2:
                    vobj.DisplayHeight = sectionplane[2]
        else:
            dim = ifc_export.get_dimension(ifcentity)
            if dim and len(dim) >= 3:
                if "Start" not in obj.PropertiesList:
                    obj.addProperty("App::PropertyVectorDistance", "Start", "Base", locked=True)
                if "End" not in obj.PropertiesList:
                    obj.addProperty("App::PropertyVectorDistance", "End", "Base", locked=True)
                if "Dimline" not in obj.PropertiesList:
                    obj.addProperty("App::PropertyVectorDistance", "Dimline", "Base", locked=True)
                obj.Start = dim[1]
                obj.End = dim[2]
                if len(dim) > 3:
                    obj.Dimline = dim[3]
                else:
                    mid = obj.End.sub(obj.Start)
                    mid.multiply(0.5)
                    obj.Dimline = obj.Start.add(mid)
            else:
                text = ifc_export.get_text(ifcentity)
                if text:
                    if "Placement" not in obj.PropertiesList:
                        obj.addProperty("App::PropertyPlacement", "Placement", "Base", locked=True)
                    if "Text" not in obj.PropertiesList:
                        obj.addProperty("App::PropertyStringList", "Text", "Base", locked=True)
                    obj.Text = [text.Literal]
                    obj.Placement = ifc_export.get_placement(ifcentity.ObjectPlacement, ifcfile)
    elif ifcentity.is_a("IfcControl"):
        ifc_psets.show_psets(obj)

    # link Label2 and Description
    if "Description" in obj.PropertiesList and hasattr(obj, "setExpression"):
        obj.setExpression("Label2", "Description")


def remove_unused_properties(obj):
    """Remove IFC properties if they are not part of the current IFC class"""

    elt = get_ifc_element(obj)
    props = list(elt.get_info().keys())
    props[props.index("id")] = "StepId"
    props[props.index("type")] = "Class"
    for prop in obj.PropertiesList:
        if obj.getGroupOfProperty(prop) == "IFC":
            if prop not in props:
                obj.removeProperty(prop)


def get_ifc_classes(obj, baseclass):
    """Returns a list of sibling classes from a given FreeCAD object"""

    # this function can become pure IFC

    if baseclass in ("IfcProject", "IfcProjectLibrary"):
        return ("IfcProject", "IfcProjectLibrary")
    ifcfile = get_ifcfile(obj)
    if not ifcfile:
        return [baseclass]
    classes = []
    schema = ifcfile.wrapped_data.schema_name()
    schema = ifcopenshell.ifcopenshell_wrapper.schema_by_name(schema)
    declaration = schema.declaration_by_name(baseclass)
    if "StandardCase" in baseclass:
        declaration = declaration.supertype()
    if declaration.supertype():
        # include sibling classes
        classes = [sub.name() for sub in declaration.supertype().subtypes()]
        # include superclass too so one can "navigate up"
        classes.append(declaration.supertype().name())
    # also include subtypes of the current class (ex, StandardCases)
    classes.extend([sub.name() for sub in declaration.subtypes()])
    if baseclass not in classes:
        classes.append(baseclass)
    return classes


def get_ifc_element(obj, ifcfile=None):
    """Returns the corresponding IFC element of an object"""

    if not ifcfile:
        ifcfile = get_ifcfile(obj)
    if ifcfile and hasattr(obj, "StepId"):
        try:
            return ifcfile.by_id(obj.StepId)
        except RuntimeError:
            # entity not found
            pass
    return None


def has_representation(element):
    """Tells if an elements has an own representation"""

    # This function can become pure IFC

    if hasattr(element, "Representation") and element.Representation:
        return True
    return False


def filter_elements(elements, ifcfile, expand=True, spaces=False, assemblies=True):
    """Filter elements list of unwanted classes"""

    # This function can become pure IFC

    # gather decomposition if needed
    if not isinstance(elements, (list, tuple)):
        elements = [elements]
    openings = False
    if assemblies and any([e.is_a("IfcOpeningElement") for e in elements]):
        openings = True
    if expand and (len(elements) == 1):
        elem = elements[0]
        if elem.is_a("IfcSpace"):
            spaces = True
        if not has_representation(elem):
            if elem.is_a("IfcProject"):
                elements = ifcfile.by_type("IfcElement")
                elements.extend(ifcfile.by_type("IfcSite"))
            else:
                decomp = ifcopenshell.util.element.get_decomposition(elem)
                if decomp:
                    # avoid replacing elements if decomp is empty
                    elements = decomp
        else:
            if elem.Representation.Representations:
                rep = elem.Representation.Representations[0]
                if rep.Items and rep.Items[0].is_a() == "IfcPolyline" and elem.IsDecomposedBy:
                    # only use the decomposition and not the polyline
                    # happens for multilayered walls exported by VectorWorks
                    # the Polyline is the wall axis
                    # see https://github.com/yorikvanhavre/FreeCAD-NativeIFC/issues/28
                    elements = ifcopenshell.util.element.get_decomposition(elem)
    if not openings:
        # Never load feature elements by default, they can be lazy loaded
        elements = [e for e in elements if not e.is_a("IfcFeatureElement")]
    # do load spaces when required, otherwise skip computing their shapes
    if not spaces:
        elements = [e for e in elements if not e.is_a("IfcSpace")]
    # skip projects
    elements = [e for e in elements if not e.is_a("IfcProject")]
    # skip furniture for now, they can be lazy loaded probably
    elements = [e for e in elements if not e.is_a("IfcFurnishingElement")]
    return elements


def set_attribute(ifcfile, element, attribute, value):
    """Sets the value of an attribute of an IFC element"""

    # This function can become pure IFC

    def differs(val1, val2):
        if val1 == val2:
            return False
        if not val1 and not val2:
            return False
        if isinstance(val1, (tuple, list)):
            if tuple(val1) == tuple(val2):
                return False
        if val1 is None and "NOTDEFINED" in str(val2).upper():
            return False
        if val1 is None and "UNDEFINED" in str(val2).upper():
            return False
        if val2 is None and "NOTDEFINED" in str(val1).upper():
            return False
        if val2 is None and "UNDEFINED" in str(val1).upper():
            return False
        return True

    if not ifcfile or not element:
        return False
    if isinstance(value, FreeCAD.Units.Quantity):
        f = get_scale(ifcfile)
        value = value.Value * f
    if attribute == "Class":
        if value != element.is_a():
            if value and value.startswith("Ifc"):
                cmd = "root.reassign_class"
                FreeCAD.Console.PrintLog(
                    "Changing IFC class value: " + element.is_a() + " to " + str(value) + "\n"
                )
                product = api_run(cmd, ifcfile, product=element, ifc_class=value)
                # TODO fix attributes
                return product
    if attribute in ["RefLongitude", "RefLatitude"]:
        c = [int(value)]
        c.append(int((value - c[0]) * 60))
        c.append(int(((value - c[0]) * 60 - c[1]) * 60))
        c.append(int((((value - c[0]) * 60 - c[1]) * 60 - c[2]) * 1.0e6))
        value = c
    cmd = "attribute.edit_attributes"
    attribs = {attribute: value}
    if hasattr(element, attribute):
        if attribute == "Name" and getattr(element, attribute) is None and value.startswith("_"):
            # do not consider default FreeCAD names given to unnamed alements
            return False
        if differs(getattr(element, attribute, None), value):
            FreeCAD.Console.PrintLog(
                "Changing IFC attribute value of "
                + str(attribute)
                + ": "
                + str(value)
                + " (original value:"
                + str(getattr(element, attribute))
                + ")"
                + "\n"
            )
            api_run(cmd, ifcfile, product=element, attributes=attribs)
            return True
    return False


def set_colors(obj, colors):
    """Sets the given colors to an object"""

    if FreeCAD.GuiUp and colors:
        try:
            vobj = obj.ViewObject
        except ReferenceError:
            # Object was probably deleted
            return
        # ifcopenshell issues (-1,-1,-1) colors if not set
        if isinstance(colors[0], (tuple, list)):
            colors = [tuple([abs(d) for d in c]) for c in colors]
        else:
            colors = [abs(c) for c in colors]
        if hasattr(vobj, "ShapeColor"):
            # 1.0 materials
            if not isinstance(colors[0], (tuple, list)):
                colors = [colors]
            # set the first color to opaque otherwise it spoils object transparency
            if len(colors) > 1:
                # TEMP HACK: if multiple colors, set everything to opaque because it looks wrong
                colors = [color[:3] + (1.0,) for color in colors]
            sapp = []
            for color in colors:
                sapp_mat = FreeCAD.Material()
                if len(color) < 4:
                    sapp_mat.DiffuseColor = color + (1.0,)
                else:
                    sapp_mat.DiffuseColor = color[:3] + (1.0 - color[3],)
                sapp_mat.Transparency = 1.0 - color[3] if len(color) > 3 else 0.0
                sapp.append(sapp_mat)
            vobj.ShapeAppearance = sapp


def get_body_context_ids(ifcfile):
    # This function can become pure IFC

    # Facetation is to accommodate broken Revit files
    # See https://forums.buildingsmart.org/t/suggestions-on-how-to-improve-clarity\
    # -of-representation-context-usage-in-documentation/3663/6?u=moult
    body_contexts = [
        c.id()
        for c in ifcfile.by_type("IfcGeometricRepresentationSubContext")
        if c.ContextIdentifier in ["Body", "Facetation"]
    ]
    # Ideally, all representations should be in a subcontext, but some BIM apps don't do this
    # correctly, so we add main contexts too
    body_contexts.extend(
        [
            c.id()
            for c in ifcfile.by_type("IfcGeometricRepresentationContext", include_subtypes=False)
            if c.ContextType == "Model"
        ]
    )
    return body_contexts


def get_plan_contexts_ids(ifcfile):
    # This function can become pure IFC

    # Annotation is to accommodate broken Revit files
    # See https://github.com/Autodesk/revit-ifc/issues/187
    return [
        c.id()
        for c in ifcfile.by_type("IfcGeometricRepresentationContext")
        if c.ContextType in ["Plan", "Annotation"]
    ]


def get_freecad_matrix(ios_matrix):
    """Converts an IfcOpenShell matrix tuple into a FreeCAD matrix"""

    # https://github.com/IfcOpenShell/IfcOpenShell/issues/1440
    # https://pythoncvc.net/?cat=203
    # https://github.com/IfcOpenShell/IfcOpenShell/issues/4832#issuecomment-2158583873
    m_l = list()
    for i in range(3):
        if len(ios_matrix) == 16:
            # IfcOpenShell 0.8
            line = list(ios_matrix[i::4])
        else:
            # IfcOpenShell 0.7
            line = list(ios_matrix[i::3])
        line[-1] *= SCALE
        m_l.extend(line)
    return FreeCAD.Matrix(*m_l)


def get_ios_matrix(m):
    """Converts a FreeCAD placement or matrix into an IfcOpenShell matrix tuple"""

    if isinstance(m, FreeCAD.Placement):
        m = m.Matrix
    mat = [
        [m.A11, m.A12, m.A13, m.A14],
        [m.A21, m.A22, m.A23, m.A24],
        [m.A31, m.A32, m.A33, m.A34],
        [m.A41, m.A42, m.A42, m.A44],
    ]
    # apply rounding because OCCT often changes 1.0 to 0.99999999999 or something
    rmat = []
    for row in mat:
        rmat.append([round(e, ROUND) for e in row])
    return rmat


def get_scale(ifcfile):
    """Returns the scale factor to convert any file length to mm"""

    scale = ifcopenshell.util.unit.calculate_unit_scale(ifcfile)
    # the above lines yields meter -> file unit scale factor. We need mm
    return 0.001 / scale


def set_placement(obj):
    """Updates the internal IFC placement according to the object placement"""

    # This function can become pure IFC

    ifcfile = get_ifcfile(obj)
    if not ifcfile:
        print("DEBUG: No ifc file for object", obj.Label, "Aborting")
    if obj.Class in ["IfcProject", "IfcProjectLibrary"]:
        return
    element = get_ifc_element(obj)
    if not hasattr(element, "ObjectPlacement"):
        # special case: this is a grid axis, it has no placement
        if element.is_a("IfcGridAxis"):
            return set_axis_points(obj, element, ifcfile)
        # other cases of objects without ObjectPlacement?
        print("DEBUG: object without ObjectPlacement", element)
        return False
    placement = FreeCAD.Placement(obj.Placement)
    placement.Base = FreeCAD.Vector(placement.Base).multiply(get_scale(ifcfile))
    new_matrix = get_ios_matrix(placement)
    old_matrix = ifcopenshell.util.placement.get_local_placement(element.ObjectPlacement)
    # conversion from numpy array
    old_matrix = old_matrix.tolist()
    old_matrix = [[round(c, ROUND) for c in r] for r in old_matrix]
    if new_matrix != old_matrix:
        FreeCAD.Console.PrintLog(
            "IFC: placement changed for "
            + obj.Label
            + " old: "
            + str(old_matrix)
            + " new: "
            + str(new_matrix)
            + "\n"
        )
        api = "geometry.edit_object_placement"
        api_run(api, ifcfile, product=element, matrix=new_matrix, is_si=False)
        return True
    return False


def set_axis_points(obj, element, ifcfile):
    """Sets the points of an axis from placement and length"""

    if element.AxisCurve.is_a("IfcPolyline"):
        p1 = obj.Placement.Base
        p2 = obj.Placement.multVec(FreeCAD.Vector(0, obj.Length.Value, 0))
        api_run(
            "attribute.edit_attributes",
            ifcfile,
            product=element.AxisCurve.Points[0],
            attributes={"Coordinates": tuple(p1)},
        )
        api_run(
            "attribute.edit_attributes",
            ifcfile,
            product=element.AxisCurve.Points[-1],
            attributes={"Coordinates": tuple(p2)},
        )
        return True
    print("DEBUG: unhandled axis type:", element.AxisCurve.is_a())
    return False


def save_ifc(obj, filepath=None):
    """Saves the linked IFC file of a project, but does not mark it as saved"""

    if not filepath:
        if getattr(obj, "IfcFilePath", None):
            filepath = obj.IfcFilePath
    if filepath:
        ifcfile = get_ifcfile(obj)
        if not ifcfile:
            ifcfile = create_ifcfile()
        ifcfile.write(filepath)
        FreeCAD.Console.PrintMessage("Saved " + filepath + "\n")


def save(obj, filepath=None):
    """Saves the linked IFC file of a project and set its saved status"""

    save_ifc(obj, filepath)
    obj.Modified = False


def aggregate(obj, parent, mode=None):
    """Takes any FreeCAD object and aggregates it to an existing IFC object.
    Mode can be 'opening' to force-create a subtraction"""

    proj = get_project(parent)
    if not proj:
        FreeCAD.Console.PrintError("The parent object is not part of an IFC project\n")
        return
    ifcfile = get_ifcfile(proj)
    if not ifcfile:
        return
    product = None
    objecttype = None
    new = False
    stepid = getattr(obj, "StepId", None)
    if stepid:
        # obj might be dragging at this point and has no project anymore
        try:
            elem = ifcfile[stepid]
            if obj.GlobalId == elem.GlobalId:
                product = elem
        except:
            pass
    if product:
        # this object already has an associated IFC product
        # print("DEBUG:", obj.Label, "is already part of the IFC document")
        newobj = obj
    else:
        ifcclass = None
        if mode == "opening":
            ifcclass = "IfcOpeningElement"
        if ifc_export.is_annotation(obj):
            product = ifc_export.create_annotation(obj, ifcfile)
            if Draft.get_type(obj) in ["DraftText", "Text"]:
                objecttype = "text"
        elif "CreateSpreadsheet" in obj.PropertiesList:
            obj.Proxy.create_ifc(obj, ifcfile)
            newobj = obj
        else:
            product = ifc_export.create_product(obj, parent, ifcfile, ifcclass)
    if product:
        exobj = get_object(product, obj.Document)
        if exobj is None:
            shapemode = getattr(parent, "ShapeMode", DEFAULT_SHAPEMODE)
            newobj = create_object(product, obj.Document, ifcfile, shapemode, objecttype)
            new = True
        else:
            newobj = exobj
        create_relationship(obj, newobj, parent, product, ifcfile, mode)
    base = getattr(obj, "Base", None)
    if base:
        # make sure the base is used only by this object before deleting
        if base.InList != [obj]:
            base = None
    # handle layer
    if FreeCAD.GuiUp:
        import FreeCADGui

        autogroup = getattr(getattr(FreeCADGui, "draftToolBar", None), "autogroup", None)
        if autogroup is not None:
            layer = FreeCAD.ActiveDocument.getObject(autogroup)
            if hasattr(layer, "StepId"):
                ifc_layers.add_to_layer(newobj, layer)
    # aggregate dependent objects
    for child in obj.InList:
        if hasattr(child, "Host") and child.Host == obj:
            aggregate(child, newobj)
        elif hasattr(child, "Hosts") and obj in child.Hosts:
            aggregate(child, newobj)
    for child in getattr(obj, "Group", []):
        if newobj.IfcClass == "IfcGroup" and child in obj.Group:
            aggregate(child, newobj)
    delete = not (PARAMS.GetBool("KeepAggregated", False))
    if new and delete and base:
        obj.Document.removeObject(base.Name)
    label = obj.Label
    if new and delete:
        obj.Document.removeObject(obj.Name)
    if new:
        newobj.Label = label  # to avoid 001-ing the Label...
    return newobj


def deaggregate(obj, parent):
    """Removes a FreeCAD object form its parent"""

    ifcfile = get_ifcfile(obj)
    element = get_ifc_element(obj)
    if not element:
        return
    try:
        api_run("aggregate.unassign_object", ifcfile, products=[element])
    except:
        # older version of ifcopenshell
        api_run("aggregate.unassign_object", ifcfile, product=element)
    parent.Proxy.removeObject(parent, obj)


def get_ifctype(obj):
    """Returns a valid IFC type from an object"""

    if hasattr(obj, "Class"):
        if "ifc" in str(obj.Class).lower():
            return obj.Class
    if hasattr(obj, "IfcType") and obj.IfcType != "Undefined":
        return "Ifc" + obj.IfcType.replace(" ", "")
    dtype = Draft.getType(obj)
    if dtype in ["App::Part", "Part::Compound", "Array"]:
        return "IfcElementAssembly"
    if dtype in ["App::DocumentObjectGroup"]:
        return "IfcGroup"
    return "IfcBuildingElementProxy"


def get_subvolume(obj):
    """returns a subface + subvolume from a window object"""

    tempface = None
    tempobj = None
    tempshape = None
    if hasattr(obj, "Proxy") and hasattr(obj.Proxy, "getSubVolume"):
        tempshape = obj.Proxy.getSubVolume(obj)
    elif hasattr(obj, "Subvolume") and obj.Subvolume:
        tempshape = obj.Subvolume
    if tempshape:
        if len(tempshape.Faces) == 6:
            # We assume the standard output of ArchWindows
            faces = sorted(tempshape.Faces, key=lambda f: f.CenterOfMass.z)
            baseface = faces[0]
            ext = faces[-1].CenterOfMass.sub(faces[0].CenterOfMass)
            tempface = obj.Document.addObject("Part::Feature", "BaseFace")
            tempface.Shape = baseface
            tempobj = obj.Document.addObject("Part::Extrusion", "Opening")
            tempobj.Base = tempface
            tempobj.DirMode = "Custom"
            tempobj.Dir = FreeCAD.Vector(ext).normalize()
            tempobj.LengthFwd = ext.Length
        else:
            tempobj = obj.Document.addObject("Part::Feature", "Opening")
            tempobj.Shape = tempshape
    if tempobj:
        tempobj.recompute()
    return tempface, tempobj


def create_relationship(old_obj, obj, parent, element, ifcfile, mode=None):
    """Creates a relationship between an IFC object and a parent IFC object"""

    if isinstance(parent, (FreeCAD.DocumentObject, FreeCAD.Document)):
        parent_element = get_ifc_element(parent)
    else:
        parent_element = parent
    uprel = None
    # case 4: anything inside group
    if parent_element.is_a("IfcGroup"):
        # special case: adding a section plane to a grouo turns it into a drawing
        # and removes it from any containment
        if element.is_a("IfcAnnotation") and element.ObjectType == "DRAWING":
            parent.ObjectType = "DRAWING"
            try:
                api_run("spatial.unassign_container", ifcfile, products=[parent_element])
            except:
                # older version of IfcOpenShell
                api_run("spatial.unassign_container", ifcfile, product=parent_element)
        # IFC objects can be part of multiple groups but we do the FreeCAD way here
        # and remove from any previous group
        for assignment in getattr(element, "HasAssignments", []):
            if assignment.is_a("IfcRelAssignsToGroup"):
                if element in assignment.RelatedObjects:
                    oldgroup = assignment.RelatingGr
                    try:
                        api_run("group.unassign_group", ifcfile, products=[element], group=oldgroup)
                    except:
                        # older version of IfcOpenShell
                        api_run("group.unassign_group", ifcfile, product=element, group=oldgroup)
        try:
            uprel = api_run("group.assign_group", ifcfile, products=[element], group=parent_element)
        except:
            # older version of IfcOpenShell
            uprel = api_run("group.assign_group", ifcfile, product=element, group=parent_element)
    # case 1: element inside spatiual structure
    elif parent_element.is_a("IfcSpatialStructureElement") and element.is_a("IfcElement"):
        # first remove the FreeCAD object from any parent
        if old_obj:
            for old_par in old_obj.InList:
                if hasattr(old_par, "Group") and old_obj in old_par.Group:
                    old_par.Group = [o for o in old_par.Group if o != old_obj]
            try:
                uprel = api_run("spatial.unassign_container", ifcfile, products=[element])
            except:
                # older version of IfcOpenShell
                uprel = api_run("spatial.unassign_container", ifcfile, product=element)
        if element.is_a("IfcOpeningElement"):
            uprel = api_run(
                "void.add_opening",
                ifcfile,
                opening=element,
                element=parent_element,
            )
        else:
            try:
                uprel = api_run(
                    "spatial.assign_container",
                    ifcfile,
                    products=[element],
                    relating_structure=parent_element,
                )
            except:
                # older version of ifcopenshell
                uprel = api_run(
                    "spatial.assign_container",
                    ifcfile,
                    product=element,
                    relating_structure=parent_element,
                )
    # case 2: door/window inside element
    # https://standards.buildingsmart.org/IFC/RELEASE/IFC4/ADD2_TC1/HTML/annex/annex-e/wall-with-opening-and-window.htm
    elif parent_element.is_a("IfcElement") and element.is_a() in [
        "IfcDoor",
        "IfcWindow",
    ]:
        if old_obj:
            tempface, tempobj = get_subvolume(old_obj)
            if tempobj:
                opening = ifc_export.create_product(tempobj, parent, ifcfile, "IfcOpeningElement")
                set_attribute(ifcfile, opening, "Name", "Opening")
                old_obj.Document.removeObject(tempobj.Name)
                if tempface:
                    old_obj.Document.removeObject(tempface.Name)
                api_run("void.add_opening", ifcfile, opening=opening, element=parent_element)
                api_run("void.add_filling", ifcfile, opening=opening, element=element)
        # windows must also be part of a spatial container
        try:
            api_run("spatial.unassign_container", ifcfile, products=[element])
        except:
            # old version of IfcOpenShell
            api_run("spatial.unassign_container", ifcfile, product=element)
        if parent_element.ContainedInStructure:
            container = parent_element.ContainedInStructure[0].RelatingStructure
            try:
                uprel = api_run(
                    "spatial.assign_container",
                    ifcfile,
                    products=[element],
                    relating_structure=container,
                )
            except:
                # old version of IfcOpenShell
                uprel = api_run(
                    "spatial.assign_container",
                    ifcfile,
                    product=element,
                    relating_structure=container,
                )
        elif parent_element.Decomposes:
            container = parent_element.Decomposes[0].RelatingObject
            try:
                uprel = api_run(
                    "aggregate.assign_object",
                    ifcfile,
                    products=[element],
                    relating_object=container,
                )
            except:
                # older version of ifcopenshell
                uprel = api_run(
                    "aggregate.assign_object",
                    ifcfile,
                    product=element,
                    relating_object=container,
                )
    # case 4: void element
    elif (parent_element.is_a("IfcElement") and element.is_a("IfcOpeningElement")) or (
        mode == "opening"
    ):
        uprel = api_run("void.add_opening", ifcfile, opening=element, element=parent_element)
    # case 3: element aggregated inside other element
    elif element.is_a("IfcProduct"):
        try:
            api_run("aggregate.unassign_object", ifcfile, products=[element])
        except:
            # older version of ifcopenshell
            api_run("aggregate.unassign_object", ifcfile, product=element)
        try:
            uprel = api_run(
                "aggregate.assign_object",
                ifcfile,
                products=[element],
                relating_object=parent_element,
            )
        except:
            # older version of ifcopenshell
            uprel = api_run(
                "aggregate.assign_object",
                ifcfile,
                product=element,
                relating_object=parent_element,
            )
    if hasattr(parent, "Proxy") and hasattr(parent.Proxy, "addObject"):
        parent.Proxy.addObject(parent, obj)
    return uprel


def get_elem_attribs(ifcentity):
    # This function can become pure IFC

    # usually info_ifcentity = ifcentity.get_info() would de the trick
    # the above could raise an unhandled exception on corrupted ifc files
    # in IfcOpenShell
    # see https://github.com/IfcOpenShell/IfcOpenShell/issues/2811
    # thus workaround

    info_ifcentity = {"id": ifcentity.id(), "class": ifcentity.is_a()}

    # get attrib keys
    attribs = []
    for anumber in range(20):
        try:
            attr = ifcentity.attribute_name(anumber)
        except Exception:
            break
        attribs.append(attr)

    # get attrib values
    for attr in attribs:
        try:
            value = getattr(ifcentity, attr)
        except Exception as e:
            value = "Error: {}".format(e)
            print(
                "DEBUG: The entity #{} has a problem on attribute {}: {}".format(
                    ifcentity.id(), attr, e
                )
            )
        info_ifcentity[attr] = value

    return info_ifcentity


def migrate_schema(ifcfile, schema):
    """migrates a file to a new schema"""

    # This function can become pure IFC

    newfile = ifcopenshell.file(schema=schema)
    migrator = ifcopenshell.util.schema.Migrator()
    table = {}
    for entity in ifcfile:
        new_entity = migrator.migrate(entity, newfile)
        table[entity.id()] = new_entity.id()
    return newfile, table


def remove_ifc_element(obj, delete_obj=False):
    """removes the IFC data associated with an object.
    If delete_obj is True, the FreeCAD object is also deleted"""

    # This function can become pure IFC

    ifcfile = get_ifcfile(obj)
    element = get_ifc_element(obj)
    if ifcfile and element:
        api_run("root.remove_product", ifcfile, product=element)
        if delete_obj:
            obj.Document.removeObject(obj.Name)
        return True
    return False


def get_orphan_elements(ifcfile):
    """returns a list of orphan products in an ifcfile"""

    products = ifcfile.by_type("IfcProduct")
    products = [p for p in products if not p.Decomposes]
    products = [p for p in products if not getattr(p, "ContainedInStructure", [])]
    products = [p for p in products if not hasattr(p, "VoidsElements") or not p.VoidsElements]
    # add control elements
    proj = ifcfile.by_type("IfcProject")[0]
    for rel in getattr(proj, "Declares", []):
        for ctrl in getattr(rel, "RelatedDefinitions", []):
            if ctrl.is_a("IfcControl"):
                products.append(ctrl)
    groups = []
    for o in products:
        for rel in getattr(o, "HasAssignments", []):
            if rel.is_a("IfcRelAssignsToGroup"):
                g = rel.RelatingGroup
                if (g not in products) and (g not in groups):
                    groups.append(g)
    products.extend(groups)
    return products


def get_group(project, name):
    """returns a group of the given type under the given IFC project. Creates it if needed"""

    if not project:
        return None
    if hasattr(project, "Group"):
        group = project.Group
    elif hasattr(project, "Objects"):
        group = project.Objects
    else:
        group = []
    for c in group:
        if c.isDerivedFrom("App::DocumentObjectGroupPython"):
            if c.Name == name:
                return c
    if hasattr(project, "Document"):
        doc = project.Document
    else:
        doc = project
    group = add_object(doc, otype="group", oname=name)
    group.Label = name.strip("Ifc").strip("Group")
    if hasattr(project.Proxy, "addObject"):
        project.Proxy.addObject(project, group)
    return group


def load_orphans(obj):
    """loads orphan objects from the given project object"""

    if isinstance(obj, FreeCAD.DocumentObject):
        doc = obj.Document
    else:
        doc = obj
    ifcfile = get_ifcfile(obj)
    shapemode = obj.ShapeMode
    elements = get_orphan_elements(ifcfile)
    objs = []
    for element in elements:
        nobj = create_object(element, doc, ifcfile, shapemode)
        objs.append(nobj)
    processed = assign_groups(elements, ifcfile)

    # put things under project. This is important so orphan elements still can find
    # their IFC file
    rest = [o for o in objs if o not in processed]
    if rest:
        project = get_project(ifcfile)
        if isinstance(project, FreeCAD.DocumentObject):
            for o in rest:
                project.Proxy.addObject(project, o)

    # TEST: Try recomputing
    QtCore.QTimer.singleShot(0, lambda: recompute(objs))


def remove_tree(objs):
    """Removes all given objects and their children, if not used by others"""

    if not objs:
        return
    doc = objs[0].Document
    nobjs = objs
    for obj in objs:
        for child in obj.OutListRecursive:
            if child not in nobjs:
                nobjs.append(child)
    deletelist = []
    for obj in nobjs:
        for par in obj.InList:
            if par not in nobjs:
                break
        else:
            deletelist.append(obj.Name)
    for n in deletelist:
        doc.removeObject(n)


def recompute(children):
    """Temporary function to recompute objects. Some objects don't get their
    shape correctly at creation"""
    doc = None
    for c in children:
        if c:
            c.touch()
            doc = c.Document
    if doc:
        doc.recompute()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2020 Travis Apple <travisapple@gmail.com>               *
# *   Copyright (c) 2025 baidakovil <baidakovil@icloud.com>                 *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

# REFS:
# https://github.com/mrdoob/three.js/blob/master/examples/webgl_interactive_buffergeometry.html
# https://threejs.org/examples/#webgl_buffergeometry_lines
# https://forum.freecad.org/viewtopic.php?t=51245
# https://forum.freecad.org/viewtopic.php?t=29487
# https://threejs.org/examples/#webgl_raycast_sprite
#
# Params for export()
#    'colors' is of the form: {'Body': [1,0,0], 'Body001': [1,1,0], 'Body002': [1,0,1] }
#    'camera' is of the form: "PerspectiveCamera {\n  viewportMapping ADJUST_CAMERA\n  position 30.242626 -51.772324 85.63475\n  orientation -0.4146691 0.088459305 -0.90566254  4.7065201\nnearDistance 53.126431\n  farDistance 123.09125\n  aspectRatio 1\n  focalDistance 104.53851\n  heightAngle 0.78539819\n\n}"
#    The 'camera' string for the active document may be generated from: import OfflineRenderingUtils; OfflineRenderingUtils.getCamera(FreeCAD.ActiveDocument.FileName);
#
# Development reload oneliner:
# def re(): from importlib import reload;import importWebGL;reload(importWebGL);o=FreeCAD.getDocument("YourDocName");importWebGL.export([o.getObject("YourBodyName")],u"C:/path/to/your/file.htm");

## @package importWebGL
#  \ingroup ARCH
#  \brief FreeCAD WebGL Exporter
#
#  This module provides tools to export HTML files containing the
#  exported objects in WebGL format and a simple three.js-based viewer.
#  Tests are provided in src/Mod/BIM/bimtests/TestWebGLExport.py.
#  The template is provided in src/Mod/BIM/Resources/templates/webgl_export_template.html.

"""FreeCAD WebGL Exporter"""

import json
import os
import textwrap
from builtins import open as pyopen
from typing_extensions import NotRequired, TypedDict

import numpy as np

import FreeCAD
import Draft
import Mesh
import OfflineRenderingUtils
import Part
from draftutils import params

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtWidgets
    from draftutils.translate import translate
else:
    FreeCADGui = None

    def translate(ctxt, txt):
        return txt


disableCompression = False  # Compress object data before sending to JS
base = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890!#$%&()*+-:;/=>?@[]^_,.{|}~`"  # safe str chars for js in all cases
baseFloat = ",.-0123456789"
threejs_version = "0.172.0"


def getHTMLTemplate():
    """Returns the HTML template from external file.
    The custom template path can be set in the Preferences.
    Returns None if no valid template is found.
    """

    def try_read_template(path, description):
        """Helper function to safely read a template file."""
        if not os.path.isfile(path):
            FreeCAD.Console.PrintWarning(
                f"{description.capitalize()} file '{path}' does not " "exist or is not a file.\n"
            )
            return None
        try:
            with open(path, "r", encoding="utf-8") as f:
                return f.read()
        except Exception as e:
            FreeCAD.Console.PrintWarning(
                f"Failed to read {description} file '{path}' " f"due to: {str(e)}\n"
            )
            return None

    using_custom_template = params.get_param("useCustomWebGLExportTemplate", path="Mod/BIM")

    # Try to use custom template if enabled
    if using_custom_template:
        custom_path = params.get_param("WebGLTemplateCustomPath", path="Mod/BIM")
        custom_content = try_read_template(custom_path, "custom WebGL template")
        if custom_content:
            FreeCAD.Console.PrintMessage(f"Using custom template file '{custom_path}'.\n")
            return custom_content
        else:
            # Custom template failed - ask user or auto-fallback
            if not FreeCADGui:
                # In non-GUI mode, cancel export when custom template fails
                FreeCAD.Console.PrintError(
                    f"Export cancelled: Custom template '{custom_path}' " "not available.\n"
                )
                return None

            # In GUI mode, ask the user
            message = translate(
                "BIM",
                "Custom WebGL template file '{}' could not be read.\n\n"
                "Do you want to proceed using the default template?",
            ).format(custom_path)

            reply = QtWidgets.QMessageBox.question(
                FreeCADGui.getMainWindow(),
                translate("BIM", "WebGL Template Not Found"),
                message,
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.Yes,
            )

            if reply != QtWidgets.QMessageBox.Yes:
                # User chose not to proceed - return None to indicate failure
                FreeCAD.Console.PrintError(
                    f"Export cancelled: Custom template '{custom_path}' " "not available.\n"
                )
                return None

    # Try to use default template
    default_template_path = os.path.join(
        FreeCAD.getResourceDir(),
        "Mod",
        "BIM",
        "Resources",
        "templates",
        "webgl_export_template.html",
    )
    default_content = try_read_template(default_template_path, "default WebGL template")
    if default_content:
        FreeCAD.Console.PrintMessage(f"Using default template file '{default_template_path}'.\n")
        return default_content

    # No template available - export cannot proceed
    if FreeCADGui:
        # In GUI mode, show Qt error dialog
        message = translate(
            "BIM",
            "The default WebGL export template is not available at path:"
            " {}\n\nPlease check your FreeCAD installation or provide a "
            "custom template under menu Preferences → Import-Export → WebGL.",
        ).format(default_template_path)

        # Use getMainWindow() as parent following FreeCAD patterns
        parent = FreeCADGui.getMainWindow()
        title = translate("BIM", "WebGL Export Template Error")

        QtWidgets.QMessageBox.critical(parent, title, message)
    else:
        # In headless mode, print to console
        FreeCAD.Console.PrintError(
            "Default WebGL export template not available at" f"path: {default_template_path}\n"
        )

    return None


def export(
    exportList, filename: str, colors: dict[str, str] | None = None, camera: str | None = None
) -> bool:
    """Exports objects to an html file.

    Returns:
        bool: True if export was successful, False if not (particularly,
        False if no template was available).
    """

    # Check template availability first, before any processing
    html = getHTMLTemplate()
    if html is None:
        # No template available - export failed
        return False

    global disableCompression, base, baseFloat

    data = {"camera": {}, "file": {}, "objects": []}

    populate_camera(data["camera"], camera)

    # Take the objects out of groups
    objectslist = Draft.get_group_contents(exportList, walls=True, addgroups=False)
    # objectslist = Arch.pruneIncluded(objectslist)

    for obj in objectslist:
        # Pull all obj data before we dig down the links
        label = obj.Label
        color, opacity = get_view_properties(obj, label, colors)

        validObject = False
        if obj.isDerivedFrom("Mesh::Feature"):
            mesh = obj.Mesh
            validObject = True
        if obj.isDerivedFrom("Part::Feature"):
            objShape = obj.Shape
            validObject = True
        if obj.isDerivedFrom("App::Link"):
            linkPlacement = obj.LinkPlacement
            while True:  # drill down to get to the actual obj
                if obj.isDerivedFrom("App::Link"):
                    if obj.ViewObject.OverrideMaterial:
                        color = Draft.getrgb(
                            obj.ViewObject.ShapeMaterial.DiffuseColor, testbw=False
                        )
                    obj = obj.LinkedObject
                    if hasattr(obj, "__len__"):
                        FreeCAD.Console.PrintMessage(f"{label}: Sub-Links are Unsupported.\n")
                        break
                elif obj.isDerivedFrom("Part::Feature"):
                    objShape = obj.Shape.copy(False)
                    objShape.Placement = linkPlacement
                    validObject = True
                    break
                elif obj.isDerivedFrom("Mesh::Feature"):
                    mesh = obj.Mesh.copy()
                    mesh.Placement = linkPlacement
                    validObject = True
                    break

        if not validObject:
            continue

        objdata = {
            "name": label,
            "color": color,
            "opacity": opacity,
            "verts": "",
            "facets": "",
            "wires": [],
            "faceColors": [],
            "facesToFacets": [],
            "floats": [],
        }

        if obj.isDerivedFrom("Part::Feature"):
            deviation = 0.5
            if FreeCADGui and hasattr(obj.ViewObject, "Deviation"):
                deviation = obj.ViewObject.Deviation

                # obj.ViewObject.DiffuseColor is length=1 when all faces are the same color, length=len(faces) for when they're not
                if len(obj.ViewObject.DiffuseColor) == len(objShape.Faces):
                    for fc in obj.ViewObject.DiffuseColor:
                        objdata["faceColors"].append(Draft.getrgb(fc, testbw=False))

            # get verts and facets for ENTIRE object
            shapeData = objShape.tessellate(deviation)
            mesh = Mesh.Mesh(shapeData)

            if len(objShape.Faces) > 1:
                # Map each Facet created by tessellate() to a Face so that it can be colored correctly using faceColors
                # This is done by matching the results of a tessellate() on EACH FACE to the overall tessellate stored in shapeData
                # if there is any error in matching these two then we display the whole object as one face and forgo the face colors
                for f in objShape.Faces:
                    faceData = f.tessellate(deviation)
                    found = True
                    # face verts. List of type Vector()
                    for fv in range(len(faceData[0])):
                        found = False
                        for sv in range(len(shapeData[0])):  # shape verts
                            # do not use isEqual() here
                            if faceData[0][fv] == shapeData[0][sv]:
                                # replace with the index of shapeData[0]
                                faceData[0][fv] = sv
                                found = True
                                break
                        if not found:
                            break
                    if not found:
                        FreeCAD.Console.PrintMessage("Facet to Face Mismatch.\n")
                        objdata["facesToFacets"] = []
                        break

                    # map each of the face facets to the shape facets and make a list of shape facet indices that belong to this face
                    facetList = []
                    for ff in faceData[1]:  # face facets
                        found = False
                        for sf in range(len(shapeData[1])):  # shape facets
                            if (
                                faceData[0][ff[0]] in shapeData[1][sf]
                                and faceData[0][ff[1]] in shapeData[1][sf]
                                and faceData[0][ff[2]] in shapeData[1][sf]
                            ):
                                facetList.append(sf)
                                found = True
                                break
                        if not found:
                            break
                    if not found:
                        FreeCAD.Console.PrintMessage("Facet List Mismatch.\n")
                        objdata["facesToFacets"] = []
                        break

                    if not disableCompression:
                        facetList = baseEncode(facetList)

                    objdata["facesToFacets"].append(facetList)

            wires = []  # Add wires
            for f in objShape.Faces:
                for w in f.Wires:
                    wo = Part.Wire(Part.__sortEdges__(w.Edges))
                    # use strings to avoid 0.00001 written as 1e-05
                    wire = []
                    for v in wo.discretize(QuasiDeflection=0.005):
                        wire.extend([f"{v.x:.5f}", f"{v.y:.5f}", f"{v.z:.5f}"])
                    wires.append(wire)

            if not disableCompression:
                wires, objdata["floats"] = compress_wires(wires, objdata["floats"])
            objdata["wires"] = wires

        vIndex = {}
        verts = []
        for p in mesh.Points:
            vIndex[p.Index] = p.Index
            verts.extend([f"{p.Vector.x:.5f}", f"{p.Vector.y:.5f}", f"{p.Vector.z:.5f}"])

        facets = [vIndex[i] for f in mesh.Facets for i in f.PointIndices]

        if not disableCompression:
            verts, objdata["floats"] = compress_verts(verts, objdata["floats"])
            objdata["floats"] = compress_floats(objdata["floats"])
            facets = baseEncode(facets)
            verts = baseEncode(verts)

        objdata["facets"] = facets
        objdata["verts"] = verts

        data["objects"].append(objdata)

    html = html.replace("$pagetitle", FreeCAD.ActiveDocument.Label)
    version = FreeCAD.Version()
    html = html.replace("$version", f"{version[0]}.{version[1]}.{version[2]}")

    # Remove data compression in JS
    data["compressed"] = not disableCompression
    data["base"] = base
    data["baseFloat"] = baseFloat

    html = html.replace("$data", json.dumps(data, separators=(",", ":")))  # Shape Data
    html = html.replace("$threejs_version", threejs_version)

    with pyopen(filename, "w", encoding="utf-8") as outfile:
        outfile.write(html)
    FreeCAD.Console.PrintMessage(translate("Arch", "Successfully written") + f" {filename}\n")
    return True


def get_view_properties(obj, label: str, colors: dict[str, str] | None) -> tuple[str, float]:
    """Get the color and opacity of the object"""
    color = "#cccccc"
    opacity = 1.0
    if FreeCADGui and hasattr(obj.ViewObject, "ShapeColor"):
        color = Draft.getrgb(obj.ViewObject.ShapeColor, testbw=False)
        opacity = int((100 - obj.ViewObject.Transparency) / 5) / 20  # 0>>1 with step of 0.05
    elif colors:
        if label in colors:
            color = Draft.getrgb(colors[label], testbw=False)
    return color, opacity


class CameraDict(TypedDict):
    """Dictionary for camera contents"""

    type: NotRequired[str]
    focalDistance: NotRequired[str]
    position_x: NotRequired[str]
    position_y: NotRequired[str]
    position_z: NotRequired[str]


def populate_camera(data: CameraDict, camera: str | None):
    if not FreeCADGui and not camera:
        camera = OfflineRenderingUtils.getCamera(FreeCAD.ActiveDocument.FileName)

    if camera:
        # REF: src/Mod/BIM/OfflineRenderingUtils.py
        camnode = OfflineRenderingUtils.getCoinCamera(camera)
        cameraPosition = camnode.position.getValue().getValue()
        data["type"] = "Orthographic"
        if "PerspectiveCamera" in camera:
            data["type"] = "Perspective"
        data["focalDistance"] = camnode.focalDistance.getValue()
        data["position_x"] = cameraPosition[0]
        data["position_y"] = cameraPosition[1]
        data["position_z"] = cameraPosition[2]
    else:
        v = FreeCADGui.ActiveDocument.ActiveView
        data["type"] = v.getCameraType()
        data["focalDistance"] = v.getCameraNode().focalDistance.getValue()
        data["position_x"] = v.viewPosition().Base.x
        data["position_y"] = v.viewPosition().Base.y
        data["position_z"] = v.viewPosition().Base.z


def compress_floats(floats: list[str]) -> str:
    """Compress floats to base 90

    Use ratio of 7x base13 to 4x base90 because 13^7 ~ 90^4
    """
    fullstr = json.dumps(floats, separators=(",", ":"))
    fullstr = fullstr.replace("[", "").replace("]", "").replace('"', "")
    floatStr = ""
    baseFloatCt = len(baseFloat)
    baseCt = len(base)
    for fs in range(0, len(fullstr), 7):  # chunks of 7 chars, skip the first one
        str7 = fullstr[fs : (fs + 7)]
        quotient = 0
        for s in range(len(str7)):
            quotient += baseFloat.find(str7[s]) * pow(baseFloatCt, (6 - s))
        for v in range(4):
            floatStr += base[quotient % baseCt]
            quotient = int(quotient / baseCt)
    return floatStr


def compress_wires(wires: list[list[str]], floats: list[str]) -> tuple[list[list[str]], list[str]]:
    """
    Create floats list to compress wires being written into the JS
    """
    lengths = []
    for w in wires:
        lengths.append(len(w))
        floats.extend(w)

    float_arr, all_wires = np.unique(floats, return_inverse=True)
    wire_arrays = np.array_split(all_wires, np.cumsum(lengths[:-1]))
    return [baseEncode(w.tolist()) for w in wire_arrays], float_arr.tolist()


def compress_verts(verts: list[str], floats: list[str]) -> tuple[list[int], list[str]]:
    """
    Create floats list to compress verts and wires being written into the JS
    """
    floats_v, ind, verts_v = np.unique(verts, return_index=True, return_inverse=True)

    # Reorder as np.unique orders the resulting array (needed for facet matching)
    floats_v = floats_v[ind.argsort()]
    reindex = dict(zip(ind.argsort(), np.arange(ind.size)))
    verts_v = np.vectorize(lambda entry: reindex[entry])(verts_v)

    # Get repeated indexes already existing from previous steps
    v_in_w = np.nonzero(np.isin(floats_v, floats))[0]
    w_in_v = np.nonzero(np.isin(floats, floats_v))[0]
    v_in_w2 = np.where(~np.isin(floats_v, floats))

    # Order values the same
    v_in_w = v_in_w[floats_v[v_in_w].argsort()]
    w_in_v = w_in_v[np.array(floats)[w_in_v].argsort()]

    # Replace repeated indexes that exist in floats
    new_index = len(floats)
    verts_v += new_index
    for vw, wv in zip(v_in_w + new_index, w_in_v):
        verts_v[verts_v == vw] = wv

    # Remove indexes of repeated entries in floats_v
    for vw in (v_in_w + new_index)[v_in_w.argsort()][::-1]:
        verts_v[verts_v > vw] -= 1

    return verts_v.tolist(), np.concatenate([floats, floats_v[v_in_w2]]).tolist()


def baseEncode(arr: list[int]) -> str:
    """Compresses an array of ints into a base90 string"""

    global base
    if len(arr) == 0:
        return ""

    longest = 0
    output = []
    baseCt = len(base)
    for v in range(len(arr)):
        buffer = ""
        quotient = arr[v]
        while True:
            buffer += base[quotient % baseCt]
            quotient = int(quotient / baseCt)
            if quotient == 0:
                break
        output.append(buffer)
        if len(buffer) > longest:
            longest = len(buffer)
    output = [("{:>" + str(longest) + "}").format(x) for x in output]  # pad each element
    return str(longest) + ("").join(output)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import pathlib
from typing import Optional, List, Type, Iterable
from PySide.QtWidgets import QFileDialog, QMessageBox
from ...assets import AssetSerializer
from ...assets.ui.util import (
    make_import_filters,
    get_serializer_from_extension,
)
from ..models.base import ToolBit
from ..serializers import all_serializers


class ToolBitOpenDialog(QFileDialog):
    def __init__(
        self,
        serializers: Iterable[Type[AssetSerializer]] | None,
        parent=None,
    ):
        super().__init__(parent)
        self.serializers = list(serializers) if serializers else all_serializers
        self.setWindowTitle("Open ToolBit(s)")
        self.setFileMode(QFileDialog.ExistingFiles)  # Allow multiple files
        filters = make_import_filters(self.serializers)
        self.setNameFilters(filters)
        if filters:
            self.selectNameFilter(filters[0])

    def _deserialize_selected_file(self, file_path: pathlib.Path) -> Optional[ToolBit]:
        """Deserialize the selected file using the appropriate serializer."""
        file_extension = file_path.suffix.lower()
        serializer_class = get_serializer_from_extension(
            self.serializers, file_extension, for_import=True
        )
        if not serializer_class:
            QMessageBox.critical(
                self,
                "Error",
                f"No supported serializer found for file extension '{file_extension}'",
            )
            return None
        try:
            raw_data = file_path.read_bytes()
            toolbit = serializer_class.deep_deserialize(raw_data)
            if not isinstance(toolbit, ToolBit):
                raise TypeError("Deserialized asset is not of type ToolBit")
            return toolbit
        except Exception as e:
            QMessageBox.critical(self, "Error", f"Failed to import toolbit: {e}")
            return None

    def exec(self) -> List[ToolBit]:
        toolbits = []
        if super().exec_():
            filenames = self.selectedFiles()
            for filename in filenames:
                file_path = pathlib.Path(filename)
                toolbit = self._deserialize_selected_file(file_path)
                if toolbit:
                    toolbits.append(toolbit)
        return toolbits



"""
this test will FAIL on current main branch (with PR #26963) and PASS after the fix.
current main at: https://github.com/FreeCAD/FreeCAD/tree/24f0c8e2c321bb202410feae84eb58779ba8e8d2
"""

import unittest
import FreeCAD as App
import Part
import Draft


class TestPartMirroringRegression(unittest.TestCase):
    """Regression test for GitHub issue #27365.

    Part::Mirroring with Draft Clone produces incorrect results after PR #26963.
    The mirror position shifts on recompute when the source has non-identity Placement.
    """

    def setUp(self):
        self.doc = App.newDocument("TestMirrorRegression")

    def tearDown(self):
        App.closeDocument(self.doc.Name)

    def testMirroringWithDraftCloneStability(self):
        """Test that Part::Mirroring position is stable across recomputes.

        This test reproduces issue #27365: mirroring a Draft Clone that has
        placement, and causes the mirror position to be incorrect and/or shift
        on recompute.

        This test FAILS on current main (after PR #26963) and should pass after a fix is applied.
        """
        # create sketch at origin
        sketch = self.doc.addObject("Sketcher::SketchObject", "Sketch")
        sketch.addGeometry(Part.LineSegment(App.Vector(0, 0, 0), App.Vector(10, 0, 0)), False)
        sketch.addGeometry(Part.LineSegment(App.Vector(10, 0, 0), App.Vector(10, 10, 0)), False)
        sketch.addGeometry(Part.LineSegment(App.Vector(10, 10, 0), App.Vector(0, 10, 0)), False)
        sketch.addGeometry(Part.LineSegment(App.Vector(0, 10, 0), App.Vector(0, 0, 0)), False)
        self.doc.recompute()

        # create draft clone at X=30 with 90° rotation around Z
        clone = Draft.make_clone(sketch)
        clone.Placement = App.Placement(App.Vector(30, 0, 0), App.Rotation(App.Vector(0, 0, 1), 90))
        self.doc.recompute()

        # verify clone is positioned correctly
        clone_bbox = clone.Shape.BoundBox
        clone_center_x = (clone_bbox.XMin + clone_bbox.XMax) / 2
        self.assertAlmostEqual(
            clone_center_x,
            30.0,
            delta=5.0,
            msg=f"clone should be centered around X=30, but is at X={clone_center_x:.2f}",
        )

        # mirror across YZ plane (X=0, normal in +X direction)
        mirror = self.doc.addObject("Part::Mirroring", "Mirror")
        mirror.Source = clone
        mirror.Base = App.Vector(0, 0, 0)
        mirror.Normal = App.Vector(1, 0, 0)
        self.doc.recompute()

        # get initial mirror position
        initial_bbox = mirror.Shape.BoundBox
        initial_center_x = (initial_bbox.XMin + initial_bbox.XMax) / 2

        # mirror should be on opposite side of plane from clone
        # clone is at X≈30, plane at X=0, so mirror should be at X≈-30
        self.assertLess(
            initial_center_x,
            -20.0,
            msg=f"mirror should be at X≈-30 (opposite side of X=0 from clone at X≈30), "
            f"but is at X={initial_center_x:.2f}",
        )

        # position must be stable across recompute
        # this is the core bug, ie. position shifts on recompute
        for i in range(5):
            self.doc.recompute()

        final_bbox = mirror.Shape.BoundBox
        final_center_x = (final_bbox.XMin + final_bbox.XMax) / 2

        # position should not change (within tolerance)
        self.assertAlmostEqual(
            initial_center_x,
            final_center_x,
            places=3,
            msg=f"mirror position shifted on recompute: "
            f"X={initial_center_x:.6f} -> X={final_center_x:.6f}",
        )


# for standalone execution
if __name__ == "__main__":
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestPartMirroringRegression))
    runner = unittest.TextTestRunner()
    runner.run(suite)
